import inspect
import logging
import typing
from collections import Counter
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import pandas as pd

from hamilton import models, node
from hamilton.dev_utils.deprecation import deprecated
from hamilton.function_modifiers import base
from hamilton.function_modifiers.configuration import ConfigResolver
from hamilton.function_modifiers.delayed import resolve as delayed_resolve
from hamilton.function_modifiers.dependencies import (
    LiteralDependency,
    SingleDependency,
    UpstreamDependency,
    source,
)

logger = logging.getLogger(__name__)

"""Decorators that replace a function's execution with specified behavior"""

# Python 3.10 + has this built in, otherwise we have to define it
try:
    from types import EllipsisType
except ImportError:
    EllipsisType = type(...)


# the following are empty functions that we can compare against to ensure that @does uses an empty function
def _empty_function():
    pass


def _empty_function_with_docstring():
    """Docstring for an empty function"""
    pass


def ensure_function_empty(fn: Callable):
    """
    Ensures that a function is empty. This is strict definition -- the function must have only one line (and
    possibly a docstring), and that line must say "pass".
    """
    if fn.__code__.co_code not in {
        _empty_function.__code__.co_code,
        _empty_function_with_docstring.__code__.co_code,
    }:
        raise base.InvalidDecoratorException(
            f"Function: {fn.__name__} is not empty. Must have only one line that "
            'consists of "pass"'
        )


class does(base.NodeCreator):
    """``@does`` is a decorator that essentially allows you to run a function over all the input parameters. \
    So you can't pass any old function to ``@does``, instead the function passed has to take any amount of inputs and \
    process them all in the same way.

    .. code-block:: python

        import pandas as pd
        from hamilton.function_modifiers import does
        import internal_package_with_logic

        def sum_series(**series: pd.Series) -> pd.Series:
            '''This function takes any number of inputs and sums them all together.'''
            ...

        @does(sum_series)
        def D_XMAS_GC_WEIGHTED_BY_DAY(D_XMAS_GC_WEIGHTED_BY_DAY_1: pd.Series,
                                      D_XMAS_GC_WEIGHTED_BY_DAY_2: pd.Series) -> pd.Series:
            '''Adds D_XMAS_GC_WEIGHTED_BY_DAY_1 and D_XMAS_GC_WEIGHTED_BY_DAY_2'''
            pass

        @does(internal_package_with_logic.identity_function)
        def copy_of_x(x: pd.Series) -> pd.Series:
            '''Just returns x'''
            pass

    The example here is a function, that all that it does, is sum all the parameters together. So we can annotate it \
    with the ``@does`` decorator and pass it the ``sum_series`` function. The ``@does`` decorator is currently limited \
    to just allow functions that consist only of one argument, a generic \\*\\*kwargs.
    """

    def __init__(self, replacing_function: Callable, **argument_mapping: Union[str, List[str]]):
        """Constructor for a modifier that replaces the annotated functions functionality with something else.
        Right now this has a very strict validation requirements to make compliance with the framework easy.

        :param replacing_function: The function to replace the original function with.
        :param argument_mapping: A mapping of argument name in the replacing function to argument name in the \
        decorating function.
        """
        self.replacing_function = replacing_function
        self.argument_mapping = argument_mapping

    @staticmethod
    def map_kwargs(kwargs: Dict[str, Any], argument_mapping: Dict[str, str]) -> Dict[str, Any]:
        """Maps kwargs using the argument mapping.
        This does 2 things:
        1. Replaces all kwargs in passed_in_kwargs with their mapping
        2. Injects all defaults from the origin function signature

        :param kwargs: Keyword arguments that will be passed into a hamilton function.
        :param argument_mapping: Mapping of those arguments to a replacing function's arguments.
        :return: The new kwargs for the replacing function's arguments.
        """
        output = {**kwargs}
        for arg_mapped_to, original_arg in argument_mapping.items():
            if original_arg in kwargs and arg_mapped_to not in argument_mapping.values():
                del output[original_arg]
            # Note that if it is not there it could be a **kwarg
            output[arg_mapped_to] = kwargs[original_arg]
        return output

    @staticmethod
    def test_function_signatures_compatible(
        fn_signature: inspect.Signature,
        replace_with_signature: inspect.Signature,
        argument_mapping: Dict[str, str],
    ) -> bool:
        """Tests whether a function signature and the signature of the replacing function are compatible.

        :param fn_signature:
        :param replace_with_signature:
        :param argument_mapping:
        :return: True if they're compatible, False otherwise
        """
        # The easy (and robust) way to do this is to use the bind with a set of dummy arguments and test if it breaks.
        # This way we're not reinventing the wheel.
        SENTINEL_ARG_VALUE = ...  # does not matter as we never use it
        # We initialize as the default values, as they'll always be injected in
        dummy_param_values = {
            key: SENTINEL_ARG_VALUE
            for key, param_spec in fn_signature.parameters.items()
            if param_spec.default is not inspect.Parameter.empty
        }
        # Then we update with the dummy values. Again, replacing doesn't matter (we'll be mimicking it later)
        dummy_param_values.update({key: SENTINEL_ARG_VALUE for key in fn_signature.parameters})
        dummy_param_values = does.map_kwargs(dummy_param_values, argument_mapping)
        try:
            # Python signatures have a bind() capability which does exactly what we want to do
            # Throws a type error if it is not valid
            replace_with_signature.bind(**dummy_param_values)
        except TypeError:
            return False
        return True

    @staticmethod
    def ensure_function_signature_compatible(
        og_function: Callable,
        replacing_function: Callable,
        argument_mapping: Dict[str, str],
    ):
        """Ensures that a function signature is compatible with the replacing function, given the argument mapping

        :param og_function: Function that's getting replaced (decorated with `@does`)
        :param replacing_function: A function that gets called in its place (passed in by `@does`)
        :param argument_mapping: The mapping of arguments from fn to replace_with
        :return:
        """
        fn_parameters = inspect.signature(og_function).parameters
        invalid_fn_parameters = []
        for param_name, param_spec in fn_parameters.items():
            if param_spec.kind not in {
                inspect.Parameter.KEYWORD_ONLY,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
            }:
                invalid_fn_parameters.append(param_name)

        if invalid_fn_parameters:
            raise base.InvalidDecoratorException(
                f"Decorated function for @does (and really, all of hamilton), "
                f"can only consist of keyword-friendly arguments. "
                f"The following parameters for {og_function.__name__} are not keyword-friendly: {invalid_fn_parameters}"
            )
        if not does.test_function_signatures_compatible(
            inspect.signature(og_function),
            inspect.signature(replacing_function),
            argument_mapping,
        ):
            raise base.InvalidDecoratorException(
                f"The following function signatures are not compatible for use with @does: "
                f"{og_function.__name__} with signature {inspect.signature(og_function)} "
                f"and replacing function {replacing_function.__name__} with signature {inspect.signature(replacing_function)}. "
                f"Mapping for arguments provided was: {argument_mapping}. You can fix this by either adjusting "
                f"the signature for the replacing function *or* adjusting the mapping."
            )

    def validate(self, fn: Callable):
        """Validates that the function:
        - Is empty (we don't want to be overwriting actual code)
        - Has a compatible return type
        - Matches the function signature with the appropriate mapping
        :param fn: Function to validate
        :raises: InvalidDecoratorException
        """
        ensure_function_empty(fn)
        does.ensure_function_signature_compatible(
            fn, self.replacing_function, self.argument_mapping
        )

    def generate_nodes(self, fn: Callable, config) -> List[node.Node]:
        """Returns one node which has the replaced functionality
        :param fn: Function to decorate
        :param config: Configuration (not used in this)
        :return: A node with the function in `@does` injected,
        and the same parameters/types as the original function.
        """

        def wrapper_function(**kwargs):
            final_kwarg_values = {
                key: param_spec.default
                for key, param_spec in inspect.signature(fn).parameters.items()
                if param_spec.default is not inspect.Parameter.empty
            }
            final_kwarg_values.update(kwargs)
            final_kwarg_values = does.map_kwargs(final_kwarg_values, self.argument_mapping)
            return self.replacing_function(**final_kwarg_values)

        return [node.Node.from_fn(fn).copy_with(callabl=wrapper_function)]


def get_default_tags(fn: Callable) -> Dict[str, str]:
    """Function that encapsulates default tags on a function.

    :param fn: the function we want to create default tags for.
    :return: a dictionary with str -> str values representing the default tags.
    """
    module_name = inspect.getmodule(fn).__name__
    return {"module": module_name}


@deprecated(
    warn_starting=(1, 20, 0),
    fail_starting=(2, 0, 0),
    use_this=delayed_resolve,
    explanation="dynamic_transform has been replaced with @resolve -- a cleaner way"
    "to utilize config for resolving decorators. Note this allows you to use any"
    "existing decorators.",
    current_version=(1, 19, 0),
    migration_guide="https://hamilton.dagworks.io/en/latest/reference/decorators/",
)
class dynamic_transform(base.NodeCreator):
    def __init__(
        self,
        transform_cls: Type[models.BaseModel],
        config_param: str,
        **extra_transform_params,
    ):
        """Constructs a model. Takes in a model_cls, which has to have a parameter."""
        self.transform_cls = transform_cls
        self.config_param = config_param
        self.extra_transform_params = extra_transform_params

    def validate(self, fn: Callable):
        """Validates that the model works with the function -- ensures:
        1. function has no code
        2. function has no parameters
        3. function has series as a return type
        :param fn: Function to validate
        :raises InvalidDecoratorException if the model is not valid.
        """

        ensure_function_empty(fn)  # it has to look exactly
        signature = inspect.signature(fn)
        if not issubclass(typing.get_type_hints(fn).get("return"), pd.Series):
            raise base.InvalidDecoratorException(
                "Models must declare their return type as a pandas Series"
            )
        if len(signature.parameters) > 0:
            raise base.InvalidDecoratorException(
                "Models must have no parameters -- all are passed in through the config"
            )

    def generate_nodes(self, fn: Callable, config: Dict[str, Any] = None) -> List[node.Node]:
        if self.config_param not in config:
            raise base.InvalidDecoratorException(
                f"Configuration has no parameter: {self.config_param}. Did you define it? If so did you spell it right?"
            )
        fn_name = fn.__name__
        transform = self.transform_cls(
            config[self.config_param], fn_name, **self.extra_transform_params
        )
        return [
            node.Node(
                name=fn_name,
                typ=typing.get_type_hints(fn).get("return"),
                doc_string=fn.__doc__,
                callabl=transform.compute,
                input_types={dep: pd.Series for dep in transform.get_dependents()},
                tags=get_default_tags(fn),
            )
        ]

    def require_config(self) -> List[str]:
        """Returns the configuration parameters that this model requires

        :return: Just the one config param used by this model
        """
        return [self.config_param]


class model(dynamic_transform):
    """Model, same as a dynamic transform"""

    def __init__(self, model_cls, config_param: str, **extra_model_params):
        super(model, self).__init__(
            transform_cls=model_cls, config_param=config_param, **extra_model_params
        )


NamespaceType = Union[str, EllipsisType, None]


class Applicable:
    """Applicable is a largely internal construct that represents a function that can be applied as a node.
    A few of these function are external-facing, however (named, when, when_not, ...)"""

    def __init__(
        self,
        fn: Callable,
        args: Tuple[Union[Any, SingleDependency], ...],
        kwargs: Dict[str, Union[Any, SingleDependency]],
        _resolvers: List[ConfigResolver] = None,
        _name: Optional[str] = None,
        _namespace: Union[str, None, EllipsisType] = ...,
    ):
        """Instantiates an Applicable.

        :param fn: Function it takes in
        :param args: Args (*args) to pass to the function
        :param kwargs: Kwargs (**kwargs) to pass to the function
        :param _resolvers: Resolvers to use for the function
        :param _name: Name of the node to be created
        :param _namespace: Namespace of the node to be created -- currently only single-level namespaces are supported
        """
        self.fn = fn
        if "_name" in kwargs:
            raise ValueError("Cannot pass in _name as a kwarg")

        self.kwargs = {key: value for key, value in kwargs.items() if key != "__name"}  # TODO --
        self.args = args
        # figure out why this was showing up in two places...
        self.resolvers = _resolvers if _resolvers is not None else []
        self.name = _name
        self.namespace = _namespace

    def _with_resolvers(self, *additional_resolvers: ConfigResolver) -> "Applicable":
        """Helper function for the .when* group"""
        return Applicable(
            fn=self.fn,
            _resolvers=self.resolvers + list(additional_resolvers),
            _name=self.name,
            _namespace=self.namespace,
            args=self.args,
            kwargs=self.kwargs,
        )

    def when(self, **key_value_pairs) -> "Applicable":
        """Choose to apply this function when all of the keys in the function
        are present in the config, and the values match the values in the config.

        :param key_value_pairs: key/value pairs to match
        :return: The Applicable with this condition applied
        """
        return self._with_resolvers(ConfigResolver.when(**key_value_pairs))

    def when_not(self, **key_value_pairs) -> "Applicable":
        """Choose to apply this function when all of the keys specified
        do not match the values specified in the config.

        :param key_value_pairs: key/value pairs to match
        :return: The Applicable with this condition applied
        """
        return self._with_resolvers(ConfigResolver.when_not(**key_value_pairs))

    def when_in(self, **key_value_group_pairs: list) -> "Applicable":
        """Choose to apply this function when all the keys provided have values contained within the list of values
        specified

        :param key_value_group_pairs: key/value pairs to match
        :return:  The Applicable with this condition applied
        """
        return self._with_resolvers(ConfigResolver.when_in(**key_value_group_pairs))

    def when_not_in(self, **key_value_group_pairs: list) -> "Applicable":
        """Choose to apply this function when all the keys provided have values not contained within the list of values
        specified.

        :param key_value_group_pairs: key/value pairs to match
        :return:  The Applicable with this condition applied
        """

        return self._with_resolvers(ConfigResolver.when_not_in(**key_value_group_pairs))

    def namespaced(self, namespace: NamespaceType) -> "Applicable":
        """Add a namespace to this node. You probably don't need this -- you should look at "named" instead.

        :param namespace: Namespace to apply, can be ..., None, or a string.
        :return: The Applicable with this namespace
        """
        return Applicable(
            fn=self.fn,
            _resolvers=self.resolvers,
            _name=self.name,
            _namespace=namespace,
            args=self.args,
            kwargs=self.kwargs,
        )

    def resolves(self, config: Dict[str, Any]) -> bool:
        """Returns whether the Applicable resolves with the given config

        :param config: Configuration to check
        :return: Whether the Applicable resolves with the given config
        """
        for resolver in self.resolvers:
            if not resolver(config):
                return False
        return True

    def named(self, name: str, namespace: NamespaceType = ...) -> "Applicable":
        """Names the function application. This has the following rules:
        1. The name will be the name passed in, this is required
        2. If the namespace is `None`, then there will be no namespace
        3. If the namespace is `...`, then the namespace will be the namespace that already exists, usually the name of the
        function that this is decorating. This is an odd case -- but it helps if you have
        multiple of the same type of operations that you want to apply across different nodes,
        or in the case of a parameterization (which is not yet supported).

        :param name: Name of the node to be created
        :param namespace: Namespace of the node to be created -- currently only single-level namespaces are supported
        :return: The applicable with the new name
        """
        return Applicable(
            fn=self.fn,
            _resolvers=self.resolvers,
            _name=name if name is not None else self.name,
            _namespace=(
                None
                if namespace is None
                else (namespace if namespace is not ... else self.namespace)
            ),
            args=self.args,
            kwargs=self.kwargs,
        )

    def get_config_elements(self) -> List[str]:
        """Returns the config elements that this Applicable uses"""
        out = []
        for resolver in self.resolvers:
            out.extend(resolver.optional_config)
        return out

    def validate(self, chain_first_param: bool, allow_custom_namespace: bool):
        """Validates that the Applicable function can be applied given the
        set of args/kwargs passed in. This says that:

        1. The signature binds appropriately
        2. If we chain the first parameter, it is not present in the function

        Note that this is currently restrictive. We only support hamilton-friendly functions. Furthermore,
        this logic is slightly duplicated from `@does` above. We will be suporting more function shapes (in
        both this and `@does`) over time, and also combining the logic between the two for
        validating/binding signatures.

        :param chain_first_param: Whether we chain the first parameter
        :raises InvalidDecoratorException if the function cannot be applied
        :return:
        """
        args = ((...,) if chain_first_param else ()) + tuple(self.args)  # dummy argument at first
        sig = inspect.signature(self.fn)
        if len(sig.parameters) == 0:
            raise base.InvalidDecoratorException(
                f"Function: {self.fn.__name__} has no parameters. "
                f"You cannot apply a function with no parameters."
            )
        invalid_args = [
            item
            for item in inspect.signature(self.fn).parameters.values()
            if item.kind
            not in {
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.KEYWORD_ONLY,
            }
        ]
        if len(invalid_args) > 0:
            raise base.InvalidDecoratorException(
                f"Function: {self.fn.__name__} has invalid parameters. "
                "You cannot apply a function with parameters that are not keyword-friendly. "
                f"The following parameters are not keyword-friendly: {invalid_args}"
            )
        try:
            sig.bind(*args, **self.kwargs)
        except TypeError as e:
            raise base.InvalidDecoratorException(
                f"Function: {self.fn.__name__} cannot be applied with the following args: {self.args} "
                f"and the following kwargs: {self.kwargs}"
            ) from e
        if len(sig.parameters) == 0:
            raise base.InvalidDecoratorException(
                f"Function: {self.fn.__name__} has no parameters. "
                "You cannot apply a function with no parameters."
            )
        if self.namespace is not ... and not allow_custom_namespace:
            raise base.InvalidDecoratorException(
                "Currently, setting namespace globally inside "
                "pipe(...)/flow(...) is not compatible with setting namespace "
                "for a step(...) call."
            )
        try:
            node.Node.from_fn(self.fn)
        except ValueError as e:
            raise base.InvalidDecoratorException(
                f"Function: {self.fn.__name__} cannot be applied with the following args: {self.args} "
                f"and the following kwargs: {self.kwargs}. See documentation on pipe(), the function "
                "shapes are currently restrictive to anything with named kwargs (either kwarg-only or positional/kwarg arguments), "
                "and must be typed. If you need functions that don't have these requirements, please reach out to the Hamilton team."
                "Current workarounds are to define a wrapper function that assigns types with the proper keyword-friendly arguments."
            ) from e

    def resolve_namespace(self, default_namespace: str) -> Tuple[str, ...]:
        """Resolves the namespace -- see rules in `named` for more details.

        :param default_namespace: namespace to use as a default if we do not wish to override it
        :return: The namespace to use, as a tuple (hierarchical)
        """
        return (
            (default_namespace,)
            if self.namespace is ...
            else (self.namespace,) if self.namespace is not None else ()
        )

    def bind_function_args(
        self, current_param: Optional[str]
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Binds function arguments, given current, chained parameeter

        :param current_param: Current, chained parameter. None, if we're not chaining.
        :return: A tuple of (upstream_inputs, literal_inputs)
        """
        args_to_bind = self.args
        if current_param is not None:
            args_to_bind = (source(current_param),) + args_to_bind
        kwargs_to_bind = self.kwargs
        fn_signature = inspect.signature(self.fn)
        bound_signature = fn_signature.bind(*args_to_bind, **kwargs_to_bind)
        all_kwargs = {**bound_signature.arguments, **bound_signature.kwargs}
        upstream_inputs = {}
        literal_inputs = {}
        # TODO -- restrict to ensure that this covers *all* dependencies
        # TODO -- bind to parameters using args
        for dep, value in all_kwargs.items():
            if isinstance(value, UpstreamDependency):
                upstream_inputs[dep] = value.source
            elif isinstance(value, LiteralDependency):
                literal_inputs[dep] = value.value
            else:
                literal_inputs[dep] = value

        return upstream_inputs, literal_inputs


def step(
    fn, *args: Union[SingleDependency, Any], **kwargs: Union[SingleDependency, Any]
) -> Applicable:
    """Applies a function to for a node (or a subcomponent of a node).
    See documentation for `pipe` to see how this is used.

    :param fn: Function to use. Must be validly called as f(**kwargs), and have a 1:1 mapping of kwargs to parameters.
    :param args: Args to pass to the function -- although these cannot be variable/position-only arguments, they can be
    positional arguments. If these are not source/value, they will be converted to a value (a literal)
    :param kwargs: Kwargs to pass to the function. These can be source/value, or they can be literals. If they are literals,
    they will be converted to a value (a literal)
    :return: an applicable with the function applied
    """
    return Applicable(fn=fn, _resolvers=[], args=args, kwargs=kwargs)


class pipe(base.NodeInjector):
    """Decorator to represent a chained set of transformations. This specifically solves the "node redefinition"
    problem, and is meant to represent a pipeline of chaining/redefinitions. This is similar (and can happily be
    used in conjunction with) `pipe` in pandas. In Pyspark this is akin to the common operation of redefining a dataframe
    with new columns. While it is generally reasonable to contain these constructs within a node's function,
    you should consider `pipe` for any of the following reasons:

    1.  You want the transformations to display as nodes in the DAG, with the possibility of storing or visualizing
    the result
    2. You want to pull in functions from an external repository, and build the DAG a little more procedurally
    3. You want to use the same function multiple times, but with different parameters -- while `@does`/`@parameterize` can
    do this, this presents an easier way to do this, especially in a chain.

    To demonstrate the rules for chaining nodes, we'll be using the following example. This is
    using primitives to demonstrate, but as hamilton is just functions of any python objects, this works perfectly with
    dataframes, series, etc...


    .. code-block:: python
        :name: Simple @pipe example

        from hamilton.function_modifiers import step, pipe, value, source

        def _add_one(x: int) -> int:
            return x + 1

        def _sum(x: int, y: int) -> int:
            return x + y

        def _multiply(x: int, y: int, z: int=10) -> int:
            return x * y * z

        @pipe(
            step(_add_one),
            step(_multiply, y=2),
            step(_sum, y=value(3)),
            step(_multiply, y=source("upstream_node_to_multiply")),
        )
        def final_result(upstream_int: int) -> int:
            return upstream_int

    .. code-block:: python
        :name: Equivalent example with no @pipe, nested

        upstream_int = ... # result from upstream
        upstream_node_to_multiply = ... # result from upstream

        output = final_result(
            _multiply(
                _sum(
                    _multiply(
                        _add_one(upstream_int),
                        y=2
                    ),
                    y=3
                ),
                y=upstream_node_to_multiply
            )
        )

    .. code-block:: python
        :name: Equivalent example with no @pipe, procedural

        upstream_int = ... # result from upstream
        upstream_node_to_multiply = ... # result from upstream

        one_added = _add_one(upstream_int)
        multiplied = _multiply(one_added, y=2)
        summed = _sum(multiplied, y=3)
        multiplied_again = _multiply(summed, y=upstream_node_to_multiply)
        output = final_result(multiplied_again)


    Note that functions must have no position-only arguments (this is rare in python, but hamilton does not handle these).
    This basically means that the functions must be defined similarly to `def fn(x, y, z=10)` and not `def fn(x, y, /, z=10)`.
    In fact, all arguments must be named and "kwarg-friendly", meaning that the function can happily be called with `**kwargs`,
    where kwargs are some set of resolved upstream values. So, no `*args` are allowed, and `**kwargs` (variable keyword-only) are not
    permitted. Note that this is not a design limitation, rather an implementation detail -- if you feel like you need this, please
    reach out.

    Furthermore, the function should be typed, as a Hamilton function would be.

    One has two ways to tune the shape/implementation of the subsequent nodes:

    1. `when`/`when_not`/`when_in`/`when_not_in` -- these are used to filter the application of the function. This is valuable to reflect
        if/else conditions in the structure of the DAG, pulling it out of functions, rather than buried within the logic itself. It is functionally
        equivalent to `@config.when`.

        For instance, if you want to include a function in the chain only when a config parameter is set to a certain value, you can do:

        .. code-block:: python

            @pipe(
                step(_add_one).when(foo="bar"),
                step(_add_two, y=source("other_node_to_add").when(foo="baz"),
            )
            def final_result(upstream_int: int) -> int:
                return upstream_int

        This will only apply the first function when the config parameter `foo` is set to `bar`, and the second when it is set to `baz`.

    2. `named` -- this is used to name the node. This is useful if you want to refer to intermediate results. If this is left out,
        hamilton will automatically name the functions in a globally unique manner. The names of
        these functions will not necessarily be stable/guaranteed by the API, so if you want to refer to them, you should use `named`.
        The default namespace will always be the name of the decorated function (which will be the last node in the chain).

        `named` takes in two parameters -- required is the `name` -- this will assign the nodes with a single name and *no* global namespace.
        For instance:

        .. code-block:: python

            @pipe(
                step(_add_one).named("a"),
                step(_add_two, y=source("upstream_node")).named("b"),
            )
            def final_result(upstream_int: int) -> int:
                return upstream_int

        The above will create two nodes, `a` and `b`. `a` will be the result of `_add_one`, and `b` will be the result of `_add_two`.
        `final_result` will then be called with the output of `b`. Note that, if these are part of a namespaced operation (a subdag, in particular),
        they *will* get the same namespace as the subdag.

        The second parameter is `namespace`. This is used to specify a namespace for the node. This is useful if you want
        to either (a) ensure that the nodes are namespaced but share a common one to avoid name clashes (usual case), or (b)
        if you want a custom namespace (unusual case). To indicate a custom namespace, one need simply pass in a string.

        To indicate that a node should share a namespace with the rest of the step(...) operations in a pipe, one can pass in `...` (the ellipsis).

        .. code-block:: python
          :name: Namespaced step


            @pipe(
                step(_add_one).named("a", namespace="foo"), # foo.a
                step(_add_two, y=source("upstream_node")).named("b", namespace=...), # final_result.b
            )
            def final_result(upstream_int: int) -> int:
                return upstream_int

        Note that if you pass a namespace argument to the `pipe` function, it will set the namespace on each step operation.
        This is useful if you want to ensure that all the nodes in a pipe have a common namespace, but you want to rename them.

        .. code-block:: python
            :name: pipe with globally applied namespace

            @pipe(
                step(_add_one).named("a"), # a
                step(_add_two, y=source("upstream_node")).named("b"), # foo.b
                namespace=..., # default -- final_result.a and final_result.b, OR
                namespace=None, # no namespace -- a and b are exposed as that, OR
                namespace="foo", # foo.a and foo.b
            )
            def final_result(upstream_int: int) -> int:
                return upstream_int

        In all likelihood, you should not be using this, and this is only here in case you want to expose a node for
        consumption/output later. Setting the namespace in individual nodes as well as in `pipe` is not yet supported.
    """

    def __init__(
        self,
        *transforms: Applicable,
        namespace: NamespaceType = ...,
        collapse=False,
        _chain=False,
    ):
        """Instantiates a `@pipe` decorator.

        :param transforms: step transformations to be applied, in order
        :param namespace: namespace to apply to all nodes in the pipe. This can be "..." (the default), which resolves to the name of the decorated function, None (which means no namespace), or a string (which means that all nodes will be namespaced with that string). Note that you can either use this *or* namespaces inside pipe()...
        :param collapse: Whether to collapse this into a single node. This is not currently supported.
        :param _chain: Whether to chain the first parameter. This is the only mode that is supported. Furthermore, this is not externally exposed. @flow will make use of this.
        """
        self.transforms = transforms
        self.collapse = collapse
        self.chain = _chain
        self.namespace = namespace

        if self.collapse:
            raise NotImplementedError(
                "Collapsing step() functions as one node is not yet implemented for pipe(). Please reach out if you want this feature."
            )

        if self.chain:
            raise NotImplementedError("@flow() is not yet supported -- this is ")

    def inject_nodes(
        self, params: Dict[str, Type[Type]], config: Dict[str, Any], fn: Callable
    ) -> Tuple[List[node.Node], Dict[str, str]]:
        """Injects nodes into the graph. This creates a node for each pipe() step,
        then reassigns the inputs to pass it in."""
        sig = inspect.signature(fn)
        first_parameter = list(sig.parameters.values())[0].name
        # use the name of the parameter to determine the first node
        # Then wire them all through in order
        # if it resolves, great
        # if not, skip that, pointing to the previous
        # Create a node along the way
        if first_parameter not in params:
            raise base.InvalidDecoratorException(
                f"Function: {fn.__name__} has a first parameter that is not a dependency. "
                f"@pipe requires the parameter names to match the function parameters. "
                f"Thus it might not be compatible with some other decorators"
            )
        current_param = first_parameter
        fn_count = Counter()
        nodes = []
        for applicable in self.transforms:
            if self.namespace is not ...:
                applicable = applicable.namespaced(
                    namespace=self.namespace
                )  # we reassign the global namespace
            if applicable.resolves(config):
                fn_name = applicable.fn.__name__
                postfix = "" if fn_count[fn_name] == 0 else f"_{fn_count[fn_name]}"
                node_name = (
                    applicable.name
                    if applicable.name is not None
                    else f"with{('_' if not fn_name.startswith('_') else '') + fn_name}{postfix}"
                )
                raw_node = node.Node.from_fn(
                    applicable.fn,
                    f"with{('_' if not fn_name.startswith('_') else '') + fn_name}{postfix}",
                )
                node_namespace = applicable.resolve_namespace(fn.__name__)
                raw_node = raw_node.copy_with(namespace=node_namespace, name=node_name)
                # TODO -- validate that the first parameter is the right type/all the same
                fn_count[fn_name] += 1
                upstream_inputs, literal_inputs = applicable.bind_function_args(current_param)
                nodes.append(
                    raw_node.reassign_inputs(
                        input_names=upstream_inputs,
                        input_values=literal_inputs,
                    )
                )
                current_param = raw_node.name
        return nodes, {first_parameter: current_param}  # rename to ensure it all works

    def validate(self, fn: Callable):
        """Validates the the individual steps work together."""
        for applicable in self.transforms:
            applicable.validate(
                chain_first_param=True, allow_custom_namespace=self.namespace is ...
            )
        # TODO -- validate that the types match on the chain (this is de-facto done later)

    def optional_config(self) -> Dict[str, Any]:
        """Declares the optional configuration keys for this decorator.
        These are configuration keys that can be used by the decorator, but are not required.
        Along with these we have *defaults*, which we will use to pass to the config.

        :return: The optional configuration keys with defaults. Note that this will return None
        if we have no idea what they are, which bypasses the configuration filtering we use entirely.
        This is mainly for the legacy API.
        """
        out = {}
        for applicable in self.transforms:
            for resolver in applicable.resolvers:
                out.update(resolver.optional_config)
        return out


# # TODO -- implement flow!
# class flow(pipe):
#     """flow() is a more flexible, power-user version of `pipe`. The rules are largely similar, with a few key differences:
#
#     1. The first parameter is not passed through -- the user is responsible for passing all parameters into a function
#     2. The final function can depend on any of the prior functions -- it will declare those as inputs using the input parameters. These will
#     be seen as inputs, regardless of namespace.
#
#     This means that `flow` can be used to construct *any* DAG -- this is why its a power-user capability. Before we dig into some examples,
#     a quick note on when to use it:
#
#     flow is meant to procedurally specify a subcomponent of the DAG. While Hamilton encourage declarative (not procedural) DAGs, there are
#     certain cases where you may find yourself wanting to more dynamically construct a DAG, for certain subcomponents. This is where
#     `flow` comes in. As we will show later in this doc, this can be very powerful when combined with `resolve` to build configuration-driven
#     DAGs. Again, however, this is meant to be a subset of a declarative DAG -- procedurally defined subdags can help with flexibility, but
#     should not be overused.
#
#     Now, let's get to some examples:
#
#     TODO -- basic example
#
#     TODO -- example with resolve
#
#     TODO -- examples with namespacing
#     """
#
#     def __init__(self, *transforms: Applicable, collapse=False):
#         super(flow, self).__init__(*transforms, collapse=collapse, _chain=False)



import logging

from hamilton.io.default_data_loaders import DATA_ADAPTERS
from hamilton.registry import register_adapter

logger = logging.getLogger(__name__)

registered = False
# Register all the default ones
if not registered:
    logger.debug(f"Registering default data loaders: {DATA_ADAPTERS}")
    for data_loader in DATA_ADAPTERS:
        register_adapter(data_loader)

registered = True



from types import ModuleType
from typing import Any, Dict, List, Optional, Tuple, Type

import pytest

from hamilton import node
from hamilton.graph import FunctionGraph
from hamilton.lifecycle.base import (
    ASYNC_HOOK,
    ASYNC_METHOD,
    REGISTERED_ASYNC_HOOKS,
    REGISTERED_ASYNC_METHODS,
    REGISTERED_SYNC_HOOKS,
    REGISTERED_SYNC_METHODS,
    REGISTERED_SYNC_VALIDATORS,
    SYNC_HOOK,
    SYNC_METHOD,
    SYNC_VALIDATOR,
    BaseDoNodeExecute,
    BaseDoValidateInput,
    BasePostGraphExecute,
    BasePreDoAnythingHook,
    BaseValidateGraph,
    BaseValidateNode,
    InvalidLifecycleAdapter,
    LifecycleAdapterSet,
    ValidationResult,
    lifecycle,
    validate_lifecycle_adapter_function,
)

from tests.lifecycle.lifecycle_adapters_for_testing import ExtendToTrackCalls


def _valid_function_empty():
    pass


def _valid_function_returns_value() -> int:
    return 1


def _valid_function_self(self):
    pass


def _valid_function_self_kwargs(self, *, a: int, b: int) -> int:
    return a + b


@pytest.mark.parametrize(
    "fn,returns_value,specified_return_value",
    [
        (_valid_function_empty, False, None),
        (_valid_function_returns_value, True, None),
        (_valid_function_returns_value, True, int),
        (_valid_function_self, False, None),
        (_valid_function_self_kwargs, True, None),
    ],
)
def test_validate_lifecycle_adapter_function_success(
    fn, returns_value: bool, specified_return_value: Optional[Type]
):
    """Test that the lifecycle adapter function works as expected."""
    validate_lifecycle_adapter_function(
        fn, returns_value=returns_value, return_type=specified_return_value
    )


def _function_with_positional_args(a, b):
    return a + b


def _function_with_mixed_args(self, a, b=None):
    pass


def _function_with_no_return_annotation():
    return 42


def _function_with_return_annotation() -> int:
    return 42


# Test cases


@pytest.mark.parametrize(
    "fn, returns_value",
    [
        (_valid_function_empty, True),
        (_valid_function_returns_value, False),
        (_valid_function_self, True),
        (_valid_function_self_kwargs, False),
        (_function_with_positional_args, True),
        (_function_with_mixed_args, False),
        (_function_with_no_return_annotation, True),
        (_function_with_return_annotation, False),
    ],
)
def test__validate_lifecycle_adapter_function_failure(fn, returns_value: bool):
    """Test that the lifecycle adapter function fails as expected for invalid cases."""
    with pytest.raises(InvalidLifecycleAdapter):
        validate_lifecycle_adapter_function(fn, returns_value=returns_value)


def test_base_hook_decorator():
    @lifecycle.base_hook("hook_for_testing")
    class TestHook:
        def hook_for_testing(self):
            pass

    assert getattr(TestHook, SYNC_HOOK) == "hook_for_testing"
    assert "hook_for_testing" in REGISTERED_SYNC_HOOKS


def test_base_hook_decorator_async():
    @lifecycle.base_hook("async_hook_for_testing")
    class TestHookAsync:
        async def async_hook_for_testing(self):
            pass

    assert getattr(TestHookAsync, ASYNC_HOOK) == "async_hook_for_testing"
    assert "async_hook_for_testing" in REGISTERED_ASYNC_HOOKS


def test_base_method_decorator():
    @lifecycle.base_method("method_for_testing")
    class TestMethod:
        def method_for_testing(self) -> int:
            pass

    assert getattr(TestMethod, SYNC_METHOD) == "method_for_testing"
    assert "method_for_testing" in REGISTERED_SYNC_METHODS


def test_base_method_decorator_async():
    @lifecycle.base_method("async_method_for_testing")
    class TestMethodAsync:
        async def async_method_for_testing(self) -> int:
            pass

    assert getattr(TestMethodAsync, ASYNC_METHOD) == "async_method_for_testing"
    assert "async_method_for_testing" in REGISTERED_ASYNC_METHODS


def test_base_validator_decorator():
    @lifecycle.base_validator("sync_validator")
    class TestValidatorSync:
        def sync_validator(self) -> Tuple[bool, Optional[str]]:
            return True, None

    assert getattr(TestValidatorSync, SYNC_VALIDATOR) == "sync_validator"
    assert "sync_validator" in REGISTERED_SYNC_VALIDATORS


def test_lifecycle_adapter_set_with_multiple_hooks():
    class MockHook1(BasePreDoAnythingHook, ExtendToTrackCalls):
        def pre_do_anything(self):
            pass

    class MockHook2(BasePreDoAnythingHook, ExtendToTrackCalls):
        def pre_do_anything(self):
            pass

    class MockHook3(BasePostGraphExecute, ExtendToTrackCalls):
        def post_graph_execute(
            self,
            *,
            run_id: str,
            graph: FunctionGraph,
            success: bool,
            error: Optional[Exception],
            results: Optional[Dict[str, Any]],
        ):
            pass

    hook_1 = MockHook1("mock_hook_1")
    hook_2 = MockHook2("mock_hook_2")
    hook_3 = MockHook3("mock_hook_3")

    adapter_set = LifecycleAdapterSet(hook_1, hook_2, hook_3)

    assert adapter_set.does_hook("pre_do_anything", is_async=False)
    assert adapter_set.does_hook("post_graph_execute", is_async=False)
    assert not adapter_set.does_hook("post_graph_execute", is_async=True)
    assert not adapter_set.does_hook("pre_node_execute", is_async=False)
    assert not adapter_set.does_hook("pre_node_execute", is_async=True)

    adapter_set.call_all_lifecycle_hooks_sync("pre_do_anything")
    assert len(hook_1.calls) == 1
    assert len(hook_2.calls) == 1
    assert len(hook_3.calls) == 0


def test_lifecycle_adapter_set_with_single_multi_hook():
    class MockMultiHook(BasePreDoAnythingHook, BasePostGraphExecute, ExtendToTrackCalls):
        def pre_do_anything(self):
            pass

        def post_graph_execute(
            self,
            *,
            run_id: str,
            graph: "FunctionGraph",
            success: bool,
            error: Optional[Exception],
            results: Optional[Dict[str, Any]],
        ):
            pass

    multi_hook = MockMultiHook("mock_multi_hook")
    adapter_set = LifecycleAdapterSet(multi_hook)

    assert adapter_set.does_hook("pre_do_anything", is_async=False)
    assert adapter_set.does_hook("post_graph_execute", is_async=False)
    assert not adapter_set.does_hook("pre_node_execute", is_async=False)
    assert not adapter_set.does_hook("pre_node_execute", is_async=True)

    adapter_set.call_all_lifecycle_hooks_sync("pre_do_anything")
    assert len(multi_hook.calls) == 1  # two pre do anything calls


def test_lifecycle_adapter_set_with_multiple_methods():
    class MockMethod1(BaseDoNodeExecute, ExtendToTrackCalls):
        def do_node_execute(
            self,
            *,
            run_id: str,
            node_: node.Node,
            kwargs: Dict[str, Any],
            task_id: Optional[str] = None,
        ) -> Any:
            return 1

    class MockMethod2(BaseDoValidateInput, ExtendToTrackCalls):
        def do_validate_input(self, *, node_type: type, input_value: Any) -> bool:
            return False

    method_1 = MockMethod1("mock_method_1")
    method_2 = MockMethod2("mock_method_2")

    adapter_set = LifecycleAdapterSet(
        method_1,
        method_2,
    )

    assert adapter_set.does_method("do_node_execute", is_async=False)
    assert adapter_set.does_method("do_validate_input", is_async=False)

    assert (
        adapter_set.call_lifecycle_method_sync(
            "do_node_execute", run_id="a", node_=None, kwargs={}, task_id=None
        )
        == 1
    )
    assert len(method_1.calls) == 1
    adapter_set.call_lifecycle_method_sync("do_validate_input", node_type=None, input_value=None)
    assert len(method_2.calls) == 1


def test_lifecycle_adapter_set_with_single_multi_method():
    class MockMultiMethod(BaseDoNodeExecute, BaseDoValidateInput, ExtendToTrackCalls):
        def do_node_execute(
            self,
            *,
            run_id: str,
            node_: node.Node,
            kwargs: Dict[str, Any],
            task_id: Optional[str] = None,
        ) -> Any:
            return 1

        def do_validate_input(self, *, node_type: type, input_value: Any) -> bool:
            return True

    multi_hook = MockMultiMethod("mock_multi_hook")
    adapter_set = LifecycleAdapterSet(multi_hook)

    assert adapter_set.does_method("do_node_execute", is_async=False)
    assert adapter_set.does_method("do_validate_input", is_async=False)
    assert not adapter_set.does_method("do_node_execute", is_async=True)

    assert (
        adapter_set.call_lifecycle_method_sync(
            "do_node_execute", run_id="a", node_=None, kwargs={}, task_id=None
        )
        == 1
    )
    assert len(multi_hook.calls) == 1
    adapter_set.call_lifecycle_method_sync("do_validate_input", node_type=None, input_value=None)
    assert len(multi_hook.calls) == 2


def test_lifecycle_adapter_set_with_multiple_validators():
    null_adapter_set = LifecycleAdapterSet()
    assert not null_adapter_set.does_validation("validate_node")
    assert not null_adapter_set.does_validation("validate_graph")

    class MockNodeValidator(BaseValidateNode, ExtendToTrackCalls):
        def validate_node(self, *, created_node: node.Node):
            return False, "Validating node"

    class MockGraphValidator(BaseValidateGraph, ExtendToTrackCalls):
        def validate_graph(
            self, *, graph: "FunctionGraph", modules: List[ModuleType], config: Dict[str, Any]
        ) -> Tuple[bool, Optional[str]]:
            return False, "Validating graph"

    adapter_set = LifecycleAdapterSet(
        node_validator := MockNodeValidator("mock_node_validator"),
        graph_validator := MockGraphValidator("mock_graph_validator"),
    )

    assert adapter_set.does_validation("validate_node")
    assert adapter_set.does_validation("validate_graph")
    assert adapter_set.call_all_validators_sync("validate_node", created_node=None) == [
        ValidationResult(success=False, error="Validating node", validator=node_validator)
    ]

    assert adapter_set.call_all_validators_sync(
        "validate_graph", graph=None, modules=[], config={}
    ) == [ValidationResult(success=False, error="Validating graph", validator=graph_validator)]
    assert len(node_validator.calls) == 1
    assert len(graph_validator.calls) == 1


def test_lifecycle_adapter_set_with_single_multi_validator():
    null_adapter_set = LifecycleAdapterSet()
    assert not null_adapter_set.does_validation("validate_node")
    assert not null_adapter_set.does_validation("validate_graph")

    class MockMultiValidator(BaseValidateNode, BaseValidateGraph, ExtendToTrackCalls):
        def validate_node(self, *, created_node: node.Node) -> Tuple[bool, Optional[str]]:
            return False, "Validating node"

        def validate_graph(
            self, *, graph: "FunctionGraph", modules: List[ModuleType], config: Dict[str, Any]
        ) -> Tuple[bool, Optional[str]]:
            return False, "Validating graph"

    adapter_set = LifecycleAdapterSet(
        multi_validator := MockMultiValidator("mock_multi_validator"),
    )

    assert adapter_set.does_validation("validate_node")
    assert adapter_set.does_validation("validate_graph")
    assert adapter_set.call_all_validators_sync("validate_node", created_node=None) == [
        ValidationResult(success=False, error="Validating node", validator=multi_validator)
    ]

    assert adapter_set.call_all_validators_sync(
        "validate_graph", graph=None, modules=[], config={}
    ) == [ValidationResult(success=False, error="Validating graph", validator=multi_validator)]
    assert len(multi_validator.calls) == 2



from collections import Counter
from types import ModuleType
from typing import Any, Dict, List, Optional

import pytest

from hamilton import ad_hoc_utils, driver, node
from hamilton.execution.executors import SynchronousLocalTaskExecutor
from hamilton.graph import FunctionGraph
from hamilton.htypes import Collect, Parallelizable
from hamilton.lifecycle.base import (
    BaseDoNodeExecute,
    BasePostGraphConstruct,
    BasePostGraphExecute,
    BasePostNodeExecute,
    BasePostTaskExecute,
    BasePreDoAnythingHook,
    BasePreGraphExecute,
    BasePreNodeExecute,
    BasePreTaskExecute,
)
from hamilton.node import Node

from .lifecycle_adapters_for_testing import (
    ExtendToTrackCalls,
    SentinelException,
    TrackingDoNodeExecuteHook,
    TrackingPostNodeExecuteHook,
    TrackingPostTaskExecuteHook,
    TrackingPreNodeExecuteHook,
)


def _sample_driver(*lifecycle_adapters):
    def n_iters(n_iters_input: int) -> int:
        return n_iters_input

    def parallel_over(n_iters: int) -> Parallelizable[int]:
        for i in range(n_iters):
            yield i

    def processed(parallel_over: int) -> int:
        return parallel_over**2

    def more_processed(processed: int, broken: bool = False) -> int:
        if broken:
            raise SentinelException()
        return processed**2

    def collect(more_processed: Collect[int]) -> int:
        return sum(more_processed)

    def output(collect: int) -> int:
        return collect

    mod = ad_hoc_utils.create_temporary_module(
        n_iters, parallel_over, processed, more_processed, collect, output
    )
    return (
        driver.Builder()
        .with_modules(mod)
        .with_adapters(*lifecycle_adapters)
        .enable_dynamic_execution(allow_experimental_mode=True)
        .with_remote_executor(SynchronousLocalTaskExecutor())
        .build()
    )


def test_individual_pre_node_execute_hook_task_based():
    hook_name = "pre_node_execute"
    hook = TrackingPreNodeExecuteHook(name=hook_name)
    dr = _sample_driver(hook)
    dr.execute(["output"], inputs={"n_iters_input": 5})
    relevant_calls = [item for item in hook.calls if item.name == hook_name]
    assert len(relevant_calls) == 14
    nodes_executed = Counter([item.bound_kwargs["node_"].name for item in relevant_calls])
    assert nodes_executed == {
        "parallel_over": 1,
        "n_iters": 1,
        "processed": 5,
        "more_processed": 5,
        "collect": 1,
        "output": 1,
    }
    run_ids = {item.bound_kwargs["run_id"] for item in relevant_calls}
    (run_id,) = run_ids
    assert len(run_ids) == 1
    assert len(run_id) > 10  # Should be UUID(ish)...


def test_individual_post_node_execute_hook_task_based():
    hook_name = "post_node_execute"
    hook = TrackingPostNodeExecuteHook(name=hook_name)
    dr = _sample_driver(hook)
    dr.execute(["output"], inputs={"n_iters_input": 5})
    relevant_calls = [item for item in hook.calls if item.name == hook_name]
    assert len(relevant_calls) == 14
    nodes_executed = Counter([item.bound_kwargs["node_"].name for item in relevant_calls])
    assert nodes_executed == {
        "parallel_over": 1,
        "n_iters": 1,
        "processed": 5,
        "more_processed": 5,
        "collect": 1,
        "output": 1,
    }
    run_ids = {item.bound_kwargs["run_id"] for item in relevant_calls}
    (run_id,) = run_ids
    assert len(run_ids) == 1
    assert len(run_id) > len(
        "TODO -- add a run ID"
    )  # This is a bit of a funny way to test that we actually added one...
    errors = {item.bound_kwargs["error"] for item in relevant_calls}
    assert len(errors) == 1  # one error, one None
    (error,) = errors
    assert error is None
    task_ids = {item.bound_kwargs["task_id"] for item in relevant_calls}
    assert (
        len(task_ids) == 2 + 5 + 1 + 1
    )  # 2 preprocessing nodes, 5 tasks, one collect, and one output node


#
def test_individual_post_node_execute_hook_task_based_with_exception():
    hook_name = "post_node_execute"
    hook = TrackingPostNodeExecuteHook(name=hook_name)
    dr = _sample_driver(hook)
    with pytest.raises(SentinelException):
        dr.execute(["output"], inputs={"n_iters_input": 1, "broken": True})
    relevant_calls = [item for item in hook.calls if item.name == hook_name]
    assert 3 < len(relevant_calls) < 9  # 3 nodes ran at least. One failed, which was counted.
    # So max is 8, if they all failed although with this implementation it will always be 4
    nodes_executed = {item.bound_kwargs["node_"].name for item in relevant_calls}
    assert nodes_executed == {"n_iters", "parallel_over", "processed", "more_processed"}
    assert {item.bound_kwargs["success"] for item in relevant_calls} == {
        True,
        False,
    }  # 2 success, 1 failure
    errors = {item.bound_kwargs["error"] for item in relevant_calls}
    assert len(errors) == 2  # one error, one None
    task_ids = {item.bound_kwargs["task_id"] for item in relevant_calls}
    assert (
        len(task_ids) >= 3
    )  # one to calc num iters, one to calc expand, and at lease one for the inner task that failed


def test_individual_post_task_execute_hook_task_based():
    hook_name = "post_task_execute"
    lifecycle_adapter = TrackingPostTaskExecuteHook(name=hook_name)
    dr = _sample_driver(lifecycle_adapter)
    dr.execute(["output"], inputs={"n_iters_input": 5})
    relevant_calls = [item for item in lifecycle_adapter.calls if item.name == hook_name]
    assert (
        len(relevant_calls) >= 2 + 5 + 1 + 1
    )  # 2 preprocessing tasks, 5 tasks, one collect, and one output node
    # Note we have >= as we actually happen to include the input, which is suboptimal, but not worth fixing now
    assert {item.bound_kwargs["success"] for item in relevant_calls} == {True}
    assert {item.bound_kwargs["error"] for item in relevant_calls} == {None}
    assert len({item.bound_kwargs["task_id"] for item in relevant_calls}) == 10  # unique task names


def test_individual_post_task_execute_hook_with_exception():
    hook_name = "post_task_execute"
    lifecycle_adapter = TrackingPostTaskExecuteHook(name=hook_name)
    dr = _sample_driver(lifecycle_adapter)
    with pytest.raises(SentinelException):
        dr.execute(["output"], inputs={"n_iters_input": 1, "broken": True})
    task_ids = {item.bound_kwargs["task_id"] for item in lifecycle_adapter.calls}
    assert (
        len(task_ids) >= 3
    )  # one to calc num iters, one to calc expand, and at lease one for the inner task that failed
    run_ids = {item.bound_kwargs["run_id"] for item in lifecycle_adapter.calls}
    assert len(run_ids) == 1
    success = {item.bound_kwargs["success"] for item in lifecycle_adapter.calls}
    assert len(success) == 2  # some failures, some successes
    errors = {item.bound_kwargs["error"] for item in lifecycle_adapter.calls}
    assert len(errors) == 2  # one error, one None


def test_individual_do_node_execute_method_task_based():
    method_name = "do_node_execute"
    method = TrackingDoNodeExecuteHook(name=method_name, additional_value=1)
    dr = _sample_driver(method)
    res = dr.execute(["output"], inputs={"n_iters_input": 5})
    relevant_calls = [item for item in method.calls if item.name == method_name]
    # node_names = Counter([item.bound_kwargs["node_"].name for item in relevant_calls])
    node_names = [item.bound_kwargs["node_"].name for item in relevant_calls]
    assert len(node_names) == 14
    task_ids = [item.bound_kwargs["task_id"] for item in relevant_calls]
    assert len(task_ids) >= 9
    assert (
        len(relevant_calls) == 2 + 5 * 2 + 1 + 1
    )  # 2 preprocessing nodes, 5 nodesx2 tasks, one collect, and one output node
    assert res == {"output": 426}  # Result of the above, computed but not explicitly drawn out


def test_multi_hook():
    class MultiHook(
        BasePreDoAnythingHook,
        BasePostGraphConstruct,
        BasePreGraphExecute,
        BasePreTaskExecute,
        BaseDoNodeExecute,
        BasePreNodeExecute,
        BasePostNodeExecute,
        BasePostTaskExecute,
        BasePostGraphExecute,
        ExtendToTrackCalls,
    ):
        def pre_task_execute(
            self,
            run_id: str,
            task_id: str,
            nodes: List[node.Node],
            inputs: Dict[str, Any],
            overrides: Dict[str, Any],
        ):
            pass

        def do_node_execute(
            self,
            run_id: str,
            node_: node.Node,
            kwargs: Dict[str, Any],
            task_id: Optional[str] = None,
        ):
            return node_(**kwargs)

        def post_task_execute(
            self,
            run_id: str,
            task_id: str,
            nodes: List[node.Node],
            results: Optional[Dict[str, Any]],
            success: bool,
            error: Exception,
        ):
            pass

        def pre_do_anything(self):
            pass

        def post_graph_construct(
            self, graph: "FunctionGraph", modules: List[ModuleType], config: Dict[str, Any]
        ):
            pass

        def pre_graph_execute(
            self,
            run_id: str,
            graph: "FunctionGraph",
            final_vars: List[str],
            inputs: Dict[str, Any],
            overrides: Dict[str, Any],
        ):
            pass

        def pre_node_execute(
            self, run_id: str, node_: Node, kwargs: Dict[str, Any], task_id: Optional[str] = None
        ):
            pass

        def post_node_execute(
            self,
            run_id: str,
            node_: node.Node,
            kwargs: Dict[str, Any],
            success: bool,
            error: Optional[Exception],
            result: Optional[Any],
            task_id: Optional[str] = None,
        ):
            pass

        def post_graph_execute(
            self,
            run_id: str,
            graph: "FunctionGraph",
            success: bool,
            error: Optional[Exception],
            results: Optional[Dict[str, Any]],
        ):
            pass

    multi_hook = MultiHook(name="multi_hook")

    dr = _sample_driver(multi_hook)
    dr.execute(["output"], inputs={"n_iters_input": 5})
    calls = multi_hook.calls
    hook_counts = Counter([item.fn.__name__ for item in calls])
    assert hook_counts == {
        "pre_do_anything": 1,
        "post_graph_construct": 1,
        "pre_graph_execute": 1,
        "pre_task_execute": 10,
        "do_node_execute": 14,
        "pre_node_execute": 14,
        "post_node_execute": 14,
        "post_task_execute": 10,
        "post_graph_execute": 1,
    }



import pandas as pd
import pyspark.sql as ps

from hamilton.htypes import column as _
from hamilton.plugins import h_spark

IntSeries = _[pd.Series, int]
FloatSeries = _[pd.Series, float]


def a(a_raw: IntSeries) -> IntSeries:
    return a_raw + 1


def b(b_raw: IntSeries) -> IntSeries:
    return b_raw + 3


def c(c_raw: IntSeries) -> FloatSeries:
    return c_raw * 3.5


def a_times_key(a: IntSeries, key: IntSeries) -> IntSeries:
    return a * key


def b_times_key(b: IntSeries, key: IntSeries) -> IntSeries:
    return b * key


def a_plus_b_plus_c(a: IntSeries, b: IntSeries, c: FloatSeries) -> FloatSeries:
    return a + b + c


def const_1() -> float:
    return 4.3


# Placing these functions here so we don't try to read the DAG
# This tests the mixing of different types, which is *only* allowed
# inside the with_columns subdag, and not yet allowed within Hamilton
# as hamilton doesn't know that they will compile to the same nodes


def _df_2_modules():
    def d(d_raw: IntSeries) -> IntSeries:
        return d_raw + 5

    def e(e_raw: int, d: int, const_1: float) -> float:
        return e_raw + d + const_1

    def f(f_raw: int) -> float:
        return f_raw * 3.5

    def multiply_d_e_f_key(
        d: IntSeries, e: FloatSeries, f: FloatSeries, key: IntSeries
    ) -> FloatSeries:
        return d * e * f * key

    return [d, e, f, multiply_d_e_f_key]


def df_1(spark_session: ps.SparkSession) -> ps.DataFrame:
    df = pd.DataFrame.from_records(
        [
            {"key": 1, "a_raw": 1, "b_raw": 2, "c_raw": 1},
            {"key": 2, "a_raw": 4, "b_raw": 5, "c_raw": 2},
            {"key": 3, "a_raw": 7, "b_raw": 8, "c_raw": 3},
            {"key": 4, "a_raw": 10, "b_raw": 11, "c_raw": 4},
            {"key": 5, "a_raw": 13, "b_raw": 14, "c_raw": 5},
        ]
    )
    return spark_session.createDataFrame(df)


@h_spark.with_columns(
    # TODO -- have a pool (module, rather than function) that we can select *from*
    # Or just select in the processed_df?
    a,
    b,
    c,
    a_times_key,
    b_times_key,
    a_plus_b_plus_c,
    select=["a_times_key", "b_times_key", "a_plus_b_plus_c"],
    columns_to_pass=["a_raw", "b_raw", "c_raw", "key"],
)
def processed_df_1(df_1: ps.DataFrame) -> ps.DataFrame:
    return df_1.select("key", "a_times_key", "b_times_key", "a_plus_b_plus_c")


def df_2(spark_session: ps.SparkSession) -> ps.DataFrame:
    df = pd.DataFrame.from_records(
        [
            {"key": 1, "d_raw": 1, "e_raw": 2, "f_raw": 5},
            {"key": 2, "d_raw": 4, "e_raw": 5, "f_raw": 10},
            {"key": 3, "d_raw": 7, "e_raw": 8, "f_raw": 15},
            {"key": 4, "d_raw": 10, "e_raw": 11, "f_raw": 20},
        ]
    )
    return spark_session.createDataFrame(df)


@h_spark.with_columns(
    *_df_2_modules(),
    select=["multiply_d_e_f_key", "d", "e", "f"],
    columns_to_pass=["d_raw", "e_raw", "f_raw", "key"],
)
def processed_df_2_joined_df_1(df_2: ps.DataFrame, processed_df_1: ps.DataFrame) -> ps.DataFrame:
    return df_2.join(processed_df_1, processed_df_1["key"] == df_2["key"], "inner").drop(df_2.key)


def final(processed_df_2_joined_df_1: ps.DataFrame) -> pd.DataFrame:
    return processed_df_2_joined_df_1.toPandas()



import json
import logging
import os
import pickle
from functools import singledispatch
from typing import Any, Callable, Dict, Optional, Set, Type

import typing_inspect

from hamilton.base import SimplePythonGraphAdapter
from hamilton.node import Node

logger = logging.getLogger(__name__)

"""
Base SERDE functions.

Basic format is:
@singledispatch
def write_<format>(data: object, filepath: str, name: str) -> None:

@singledispatch
def read_<format>(data: object, filepath: str) -> Any:

Functions should register themselves with the appropriate type.
"""


@singledispatch
def write_feather(data: object, filepath: str, name: str) -> None:
    """Writes data to a feather file."""
    raise NotImplementedError(f"No feather writer for type {type(data)} registered.")


@singledispatch
def read_feather(data: object, filepath: str) -> Any:
    """Reads from a feather file"""
    raise NotImplementedError(f"No feather reader for type {type(data)} registered.")


@singledispatch
def write_parquet(data: object, filepath: str, name: str) -> None:
    """Writes data to a parquet file."""
    raise NotImplementedError(f"No parquet writer for type {type(data)} registered.")


@singledispatch
def read_parquet(data: object, filepath: str) -> Any:
    """Reads from a parquet file"""
    raise NotImplementedError(f"No parquet reader for type {type(data)} registered.")


@singledispatch
def write_json(data: object, filepath: str, name: str) -> None:
    """Writes data to a json file."""
    raise NotImplementedError(f"No json writer for type {type(data)} registered.")


@singledispatch
def read_json(data: object, filepath: str) -> Any:
    """Reads from a json file"""
    raise NotImplementedError(f"No json reader for type {type(data)} registered.")


@singledispatch
def write_pickle(data: Any, filepath: str, name: str) -> None:
    """Writes data to a pickle file."""
    raise NotImplementedError(f"No object writer for type {type(data)} registered.")


@singledispatch
def read_pickle(data: Any, filtepath: str) -> object:
    """Reads from a pickle file"""
    raise NotImplementedError(f"No object reader for type {type(data)} registered.")


try:
    import pandas as pd  # conditional import to avoid pandas dependency

    @write_json.register(pd.DataFrame)
    def write_json_pd1(data: pd.DataFrame, filepath: str, name: str) -> None:
        """Writes a dataframe to a feather file."""
        return data.to_json(filepath)

    @write_json.register(pd.Series)
    def write_json_pd2(data: pd.Series, filepath: str, name: str) -> None:
        """Writes a series to a feather file."""
        _df = data.to_frame(name=name)
        return _df.to_json(filepath)

    @read_json.register(pd.Series)
    def read_json_pd1(data: pd.Series, filepath: str) -> pd.Series:
        """Reads a series from a feather file."""
        _df = pd.read_json(filepath)
        return _df[_df.columns[0]]

    @read_json.register(pd.DataFrame)
    def read_json_pd2(data: pd.DataFrame, filepath: str) -> pd.DataFrame:
        """Reads a dataframe from a feather file."""
        return pd.read_json(filepath)

    try:
        import pyarrow  # noqa: F401  # conditional import to avoid pyarrow dependency

        @write_feather.register(pd.DataFrame)
        def write_feather_pd1(data: pd.DataFrame, filepath: str, name: str) -> None:
            """Writes a dataframe to a feather file."""
            data.to_feather(filepath)

        @write_feather.register(pd.Series)
        def write_feather_pd2(data: pd.Series, filepath: str, name: str) -> None:
            """Writes a series to a feather file."""
            data.to_frame(name=name).to_feather(filepath)

        @read_feather.register(pd.DataFrame)
        def read_feather_pd1(data: pd.DataFrame, filepath: str) -> pd.DataFrame:
            """Reads a dataframe from a feather file."""
            return pd.read_feather(filepath)

        @read_feather.register(pd.Series)
        def read_feather_pd2(data: pd.Series, filepath: str) -> pd.Series:
            """Reads a series from a feather file."""
            _df = pd.read_feather(filepath)
            return _df[_df.columns[0]]

        @write_parquet.register(pd.DataFrame)
        def write_parquet_pd1(data: pd.DataFrame, filepath: str, name: str) -> None:
            """Writes a dataframe to a parquet file."""
            data.to_parquet(filepath)

        @write_parquet.register(pd.Series)
        def write_parquet_pd2(data: pd.Series, filepath: str, name: str) -> None:
            """Writes a data frame to a parquet file."""
            data.to_frame(name=name).to_parquet(filepath)

        @read_parquet.register(pd.DataFrame)
        def read_parquet_pd1(data: pd.DataFrame, filepath: str) -> pd.DataFrame:
            """Reads a dataframe from a parquet file."""
            return pd.read_parquet(filepath)

        @read_parquet.register(pd.Series)
        def read_parquet_pd2(data: pd.Series, filepath: str) -> pd.Series:
            """Reads a series from a parquet file."""
            _df = pd.read_parquet(filepath)
            return _df[_df.columns[0]]

    except ImportError:
        pass


except ImportError:
    pass


@write_json.register(dict)
def write_json_dict(data: dict, filepath: str, name: str) -> None:
    """Writes a dictionary to a JSON file."""
    if isinstance(data, dict):
        with open(filepath, "w", encoding="utf8") as file:
            json.dump(data, file)
    else:
        raise ValueError(f"Expected a dict, got {type(data)}")


@read_json.register(dict)
def read_json_dict(data: dict, filepath: str) -> dict:
    """Reads a dictionary from a JSON file."""
    with open(filepath, "r", encoding="utf8") as file:
        return json.load(file)


@write_pickle.register(object)
def write_pickle_object(data: object, filepath: str, name: str) -> None:
    if isinstance(data, object):
        print(filepath)
        with open(filepath, "wb") as file:
            pickle.dump(data, file)
    else:
        raise ValueError(f"Expected an object, got {type(data)}")


@read_pickle.register(object)
def read_pickle_object(data: object, filepath: str) -> object:
    """Reads a pickle file"""
    print(filepath)
    with open(filepath, "rb") as file:
        return pickle.load(file)


class CachingGraphAdapter(SimplePythonGraphAdapter):
    """Caching adapter.

    Any node with tag "cache" will be cached (or loaded from cache) in the format defined by the
    tag's value. There are a handful of formats supported, and other formats' readers and writers
    can be provided to the constructor.

    Values are loaded from cache if the node's file exists, unless one of these is true:
     * node is explicitly forced to be computed with a constructor argument,
     * any of its (potentially transitive) dependencies that are configured to be cached
       was nevertheless computed (either forced or missing cached file).

    Custom Serializers
    ------------------

    One can provide custom readers and writers for any format by passing them to the constructor.
    These readers and writers will override the default ones. If you don't want to override, but
    rather extend the default ones, you can do so by registering them with the `register` method
    on the appropriate function.

    Writer functions need to have the following signature:
    `def write_<format>(data: Any, filepath: str, name: str) -> None: ...`
    where `data` is the data to be written, `filepath` is the path to the file to be written to,
    and `name` is the name of the node that is being written.

    Reader functions need to have the following signature:
    `def read_<format>(data: Any, filepath: str) -> Any: ...`
    where `data` is an EMPTY OBJECT of the type you wish to instantiate, and `filepath` is the
    path to the file to be read from.

    For example, if you want to extend JSON reader/writer to work with your custom type `T`,
    you can do the following:

    .. code-block:: python

        @write_json.register(T)
        def write_json_pd1(data: T, filepath: str, name: str) -> None:
            ...

        @read_json.register(T)
        def read_json_dict(data: T, filepath: str) -> T:
            ...

    Usage
    -----

    This is a simple example of the usage of `CachingGraphAdapter`.

    First, let's define some nodes in `nodes.py`:

    .. code-block:: python

        import pandas as pd
        from hamilton.function_modifiers import tag

        def data_a() -> pd.DataFrame:
            ...

        @tag(cache="parquet")
        def data_b() -> pd.DataFrame:
            ...

        def transformed(data_a: pd.DataFrame, data_b: pd.DataFrame) -> pd.DataFrame:
            ...

    Notice that `data_b` is configured to be cached in a parquet file.

    We then simply initialize the driver with a caching adapter:

    .. code-block:: python

        from hamilton import base
        from hamilton.driver import Driver
        from hamilton.experimental import h_cache

        import nodes

        adapter = h_cache.CachingGraphAdapter(cache_path, base.PandasDataFrameResult())
        dr = Driver(config, nodes, adapter=adapter)
        result = dr.execute(["transformed"])

        # Because `data_b` has been cached now, only `data_a` and `transformed` nodes
        # will actually run.
        result = dr.execute(["transformed"])
    """

    def __init__(
        self,
        cache_path: str,
        *args,
        force_compute: Optional[Set[str]] = None,
        writers: Optional[Dict[str, Callable[[Any, str, str], None]]] = None,
        readers: Optional[Dict[str, Callable[[Any, str], Any]]] = None,
        **kwargs,
    ):
        """Constructs the adapter.

        :param cache_path: Path to the directory where cached files are stored.
        :param force_compute: Set of nodes that should be forced to compute even if cache exists.
        :param writers: A dictionary of writers for custom formats.
        :param readers: A dictionary of readers for custom formats.
        """

        super().__init__(*args, **kwargs)
        self.cache_path = cache_path
        self.force_compute = force_compute if force_compute is not None else {}
        self.computed_nodes = set()

        self.writers = writers or {}
        self.readers = readers or {}

        self._init_default_readers_writers()

    def _init_default_readers_writers(self):
        if "json" not in self.writers:
            self.writers["json"] = write_json
        if "json" not in self.readers:
            self.readers["json"] = read_json

        if "feather" not in self.writers:
            self.writers["feather"] = write_feather
        if "feather" not in self.readers:
            self.readers["feather"] = read_feather

        if "parquet" not in self.writers:
            self.writers["parquet"] = write_parquet
        if "parquet" not in self.readers:
            self.readers["parquet"] = read_parquet

        if "pickle" not in self.writers:
            self.writers["pickle"] = write_pickle
        if "pickle" not in self.readers:
            self.readers["pickle"] = read_pickle

    def _check_format(self, fmt):
        if fmt not in self.writers:
            raise ValueError(f"invalid cache format: {fmt}")

    def _write_cache(self, fmt: str, data: Any, filepath: str, node_name: str) -> None:
        self._check_format(fmt)
        self.writers[fmt](data, filepath, node_name)

    def _read_cache(self, fmt: str, expected_type: Any, filepath: str) -> None:
        self._check_format(fmt)
        return self.readers[fmt](expected_type, filepath)

    def _get_empty_expected_type(self, expected_type: Type) -> Any:
        if typing_inspect.is_generic_type(expected_type):
            return typing_inspect.get_origin(expected_type)()
        return expected_type()  # This ASSUMES that we can just do `str()`, `pd.DataFrame()`, etc.

    def execute_node(self, node: Node, kwargs: Dict[str, Any]) -> Any:
        """Executes nodes conditionally according to caching rules.

        This node is executed if at least one of these is true:

        * no cache is present,
        * it is explicitly forced by passing it to the adapter in ``force_compute``,
        * at least one of its upstream nodes that had a @cache annotation was computed,
          either due to lack of cache or being explicitly forced.

        """
        cache_format = node.tags.get("cache")
        implicitly_forced = any(dep.name in self.computed_nodes for dep in node.dependencies)
        if cache_format is not None:
            filepath = f"{self.cache_path}/{node.name}.{cache_format}"
            explicitly_forced = node.name in self.force_compute
            if explicitly_forced or implicitly_forced or not os.path.exists(filepath):
                result = node.callable(**kwargs)
                logger.debug(
                    "Writing cache for %s to %s with type %s to %s",
                    node.name,
                    filepath,
                    type(result),
                    cache_format,
                )
                self._write_cache(cache_format, result, filepath, node.name)
                self.computed_nodes.add(node.name)
                return result
            empty_expected_type = self._get_empty_expected_type(node.type)
            logger.debug(
                "Reading cache for %s from %s with type %s to %s",
                node.name,
                filepath,
                type(empty_expected_type),
                cache_format,
            )
            return self._read_cache(cache_format, empty_expected_type, filepath)

        if implicitly_forced:
            # For purposes of caching, we only mark it as computed if any cached input was computed.
            # Otherwise, dependants would always be recomputed if they have a non-cached dependency.
            self.computed_nodes.add(node.name)
        return node.callable(**kwargs)

    def build_result(self, **outputs: Dict[str, Any]) -> Any:
        """Clears the computed nodes information and delegates to the super class."""
        self.computed_nodes = set()
        return super().build_result(**outputs)



from typing import Any

import numpy as np

from hamilton import registry

try:
    import vaex
except ImportError:
    raise NotImplementedError("Vaex is not installed.")

DATAFRAME_TYPE = vaex.dataframe.DataFrame
COLUMN_TYPE = vaex.expression.Expression


@registry.get_column.register(vaex.dataframe.DataFrame)
def get_column_vaex(df: vaex.dataframe.DataFrame, column_name: str) -> vaex.expression.Expression:
    return df[column_name]


@registry.fill_with_scalar.register(vaex.dataframe.DataFrame)
def fill_with_scalar_vaex(
    df: vaex.dataframe.DataFrame, column_name: str, scalar_value: Any
) -> vaex.dataframe.DataFrame:
    df[column_name] = np.full((df.shape[0],), scalar_value)
    return df


def register_types():
    """Function to register the types for this extension."""
    registry.register_types("vaex", DATAFRAME_TYPE, COLUMN_TYPE)


register_types()



import abc
import dataclasses
import typing
from typing import Any, Collection, Dict, Tuple, Type

from hamilton.htypes import custom_subclass_check


class AdapterCommon(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def applicable_types(cls) -> Collection[Type]:
        """Returns the types that this data loader can load to.
        These will be checked against the desired type to determine
        whether this is a suitable loader for that type.

        Note that a loader can load to multiple types. This is the function to
        override if you want to add a new type to a data loader.

        Note if you have any specific requirements for loading types (generic/whatnot),
        you can override applies_to as well, but it will make it much harder to document/determine
        what is happening.

        :return:
        """
        pass

    @classmethod
    @abc.abstractmethod
    def applies_to(cls, type_: Type[Type]) -> bool:
        """Tells whether or not this adapter applies to the given type.

        Note: you need to understand the edge direction to properly determine applicability.
        For loading data, the loader type needs to be a subclass of the type being loaded into.
        For saving data, the saver type needs to be a superclass of the type being passed in.

        This is a classmethod as it will be easier to validate, and we have to
        construct this, delayed, with a factory.

        :param type_: Candidate type
        :return: True if this adapter can be used with that type, False otherwise.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def name(cls) -> str:
        """Returns the name of the data loader. This is used to register the data loader
        with the load_from decorator.

        :return: The name of the data loader.
        """
        pass

    @classmethod
    def _ensure_dataclass(cls):
        if not dataclasses.is_dataclass(cls):
            raise TypeError(
                f"DataLoader subclasses must be dataclasses. {cls.__qualname__} is not."
                f" Did you forget to add @dataclass?"
            )

    @classmethod
    def get_required_arguments(cls) -> Dict[str, Type[Type]]:
        """Gives the required arguments for the class.
        Note that this just uses the type hints from the dataclass.

        :return: The required arguments for the class.
        """
        cls._ensure_dataclass()
        type_hints = typing.get_type_hints(cls)
        return {
            field.name: type_hints.get(field.name)
            for field in dataclasses.fields(cls)
            if field.default == dataclasses.MISSING
        }

    @classmethod
    def get_optional_arguments(cls) -> Dict[str, Type[Type]]:
        """Gives the optional arguments for the class.
        Note that this just uses the type hints from the dataclass.

        :return: The optional arguments for the class.
        """
        cls._ensure_dataclass()
        type_hints = typing.get_type_hints(cls)
        return {
            field.name: type_hints.get(field.name)
            for field in dataclasses.fields(cls)
            if field.default != dataclasses.MISSING
        }

    @classmethod
    def can_load(cls) -> bool:
        """Returns whether this adapter can "load" data.
        Subclasses are meant to implement this function to
        tell the framework what to do with them.

        :return:
        """
        return False

    @classmethod
    def can_save(cls) -> bool:
        """Returns whether this adapter can "save" data.
        Subclasses are meant to implement this function to
        tell the framework what to do with them.

        :return:
        """
        return False


class DataLoader(AdapterCommon, abc.ABC):
    """Base class for data loaders. Data loaders are used to load data from a data source.
    Note that they are inherently polymorphic -- they declare what type(s) they can load to,
    and may choose to load differently depending on the type they are loading to.
    """

    @abc.abstractmethod
    def load_data(self, type_: Type[Type]) -> Tuple[Type, Dict[str, Any]]:
        """Loads the data from the data source.
        Note this uses the constructor parameters to determine
        how to load the data.

        :return: The type specified
        """
        pass

    @classmethod
    def can_load(cls) -> bool:
        return True

    @classmethod
    def applies_to(cls, type_: Type[Type]) -> bool:
        """Tells whether or not this data loader can load to a specific type.
        For instance, a CSV data loader might be able to load to a dataframe,
        a json, but not an integer.

        I.e. is the adapter type a subclass of the passed in type?

        This is a classmethod as it will be easier to validate, and we have to
        construct this, delayed, with a factory.

        :param type_: Candidate type
        :return: True if this data loader can load to the type, False otherwise.
        """
        for load_to in cls.applicable_types():
            # is the adapter type `load_to` a subclass of `type_` ?
            if custom_subclass_check(load_to, type_):
                return True
        return False


class DataSaver(AdapterCommon, abc.ABC):
    """Base class for data savers. Data savers are used to save data to a data source.
    Note that they are inherently polymorphic -- they declare what type(s) they can save from,
    and may choose to save differently depending on the type they are saving from.
    """

    @abc.abstractmethod
    def save_data(self, data: Any) -> Dict[str, Any]:
        """Saves the data to the data source.
            Note this uses the constructor parameters to determine
            how to save the data.

        :return: Any relevant metadata. This is up the the data saver, but will likely
            include the URI, etc... This is going to be similar to the metadata returned
            by the data loader in the loading tuple.
        """
        pass

    @classmethod
    def can_save(cls) -> bool:
        return True

    @classmethod
    def applies_to(cls, type_: Type[Type]) -> bool:
        """Tells whether or not this data saver can ingest a specific type to save it.

        I.e. is the adapter type a superclass of the passed in type?

        This is a classmethod as it will be easier to validate, and we have to
        construct this, delayed, with a factory.

        :param type_: Candidate type
        :return: True if this data saver can handle to the type, False otherwise.
        """
        applicable_types = cls.applicable_types()
        if len(applicable_types) > 1 and typing.Union[tuple(applicable_types)] == type_:
            # if someone outputs the union of what we support we should match it.
            return True
        for save_to in applicable_types:
            # is the adapter type `save_to` a superclass of `type_` ?
            # i.e. is `type_` a subclass of `save_to` ?
            if custom_subclass_check(type_, save_to):
                return True
        return False



from pathlib import Path
from typing import List, Optional

from hamilton import ad_hoc_utils, driver
from hamilton.cli import logic


def build(modules: List[Path], context_path: Optional[Path] = None):
    """Build a Hamilton driver from the passed modules, and
    load the Driver config from the context file.

    Dynamic execution is enabled by default to support dataflow
    using Parallelizable/Collect. This only matters if we are to
    execute code.
    """
    context = logic.load_context(context_path) if context_path else {}
    module_objects = [ad_hoc_utils.module_from_source(p.read_text()) for p in modules]
    return (
        driver.Builder()
        .enable_dynamic_execution(allow_experimental_mode=True)
        .with_modules(*module_objects)
        .with_config(context.get("HAMILTON_CONFIG", {}))
        .build()
    )


def diff(
    current_dr: driver.Driver,
    modules: List[Path],
    git_reference: Optional[str] = "HEAD",
    view: bool = False,
    output_file_path: Path = Path("./diff.png"),
    context_path: Optional[Path] = None,
) -> dict:
    """Get the diff of"""
    context = logic.load_context(context_path) if context_path else {}

    current_version = logic.hash_hamilton_nodes(current_dr)
    current_node_to_func = logic.map_nodes_to_functions(current_dr)

    reference_modules = logic.load_modules_from_git(modules, git_reference)
    reference_dr = (
        driver.Builder()
        .enable_dynamic_execution(allow_experimental_mode=True)
        .with_modules(*reference_modules)
        .with_config(context.get("HAMILTON_CONFIG", {}))
        .build()
    )
    reference_version = logic.hash_hamilton_nodes(reference_dr)
    reference_node_to_func = logic.map_nodes_to_functions(reference_dr)

    nodes_diff = logic.diff_versions(
        reference_map=reference_version,
        current_map=current_version,
    )

    full_diff = dict()
    for status, node_names in nodes_diff.items():
        full_diff[status] = dict()
        for node_name in node_names:
            if current_node_to_func.get(node_name):
                func_name = current_node_to_func.get(node_name)
            else:
                func_name = reference_node_to_func.get(node_name)

            full_diff[status][node_name] = func_name

    if view:
        dot = logic.visualize_diff(current_dr=current_dr, reference_dr=reference_dr, **nodes_diff)

        # simplified logic from hamilton.graph.display()
        output_format = "png"
        if output_file_path:  # infer format from path
            if output_file_path.suffix != "":
                output_format = output_file_path.suffix.partition(".")[-1]

        dot.render(output_file_path.with_suffix(""), format=output_format)

    return full_diff


def validate(dr: driver.Driver, context_path: Path) -> dict:
    """Use driver.validate_execution() with values from the context file"""
    context = logic.load_context(context_path)

    try:
        dr.validate_execution(
            final_vars=context["HAMILTON_FINAL_VARS"],
            inputs=context["HAMILTON_INPUTS"],
            overrides=context["HAMILTON_OVERRIDES"],
        )
    except ValueError as e:
        raise e

    return context


def version(dr: driver.Driver) -> dict:
    """Get the node and dataflow versions from the instantiated Driver"""
    nodes_hash = logic.hash_hamilton_nodes(dr)
    dataflow_hash = logic.hash_dataflow(nodes_hash)
    return dict(
        nodes_hash=nodes_hash,
        dataflow_hash=dataflow_hash,
    )


def view(dr: driver.Driver, output_file_path: Path = Path("dag.png")) -> None:
    """Display all functions of the instantiated Driver"""
    dr.display_all_functions(output_file_path)



try:
    from .version import VERSION as __version__  # noqa: F401
except ImportError:
    from version import VERSION as __version__  # noqa: F401

# this supposedly is required for namespace packages to work.
__path__ = __import__("pkgutil").extend_path(__path__, __name__)



import json

from hamilton import driver, graph_types

from tests.resources.dynamic_parallelism import no_parallel

# `version` value isn't hardcoded because it's Python version-dependent
EXPECTED_JSON = {
    "nodes": [
        {
            "name": "final",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": ["sum_step_squared_plus_step_cubed"],
            "optional_dependencies": [],
            "source": (
                "def final(sum_step_squared_plus_step_cubed: int) -> int:\n"
                '    print("finalizing")\n'
                "    return sum_step_squared_plus_step_cubed\n"
            ),
            "documentation": "",
            "version": None,  # `version` value isn't hardcoded because it's Python version-dependent
        },
        {
            "name": "number_of_steps",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": [],
            "optional_dependencies": [],
            "source": "def number_of_steps() -> int:\n    return 5\n",
            "documentation": "",
            "version": None,
        },
        {
            "name": "step_cubed",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": ["steps"],
            "optional_dependencies": [],
            "source": (
                "def step_cubed(steps: int) -> int:\n"
                '    print("cubing step {}".format(steps))\n'
                "    return steps**3\n"
            ),
            "documentation": "",
            "version": None,
        },
        {
            "name": "step_squared",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": ["steps"],
            "optional_dependencies": [],
            "source": 'def step_squared(steps: int) -> int:\n    print("squaring step {}".format(steps))\n   '
            " return steps**2\n",
            "documentation": "",
            "version": None,
        },
        {
            "name": "step_squared_plus_step_cubed",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": ["step_cubed", "step_squared"],
            "optional_dependencies": [],
            "source": (
                "def step_squared_plus_step_cubed(step_squared: int, step_cubed: int) -> int:\n"
                '    print("adding step squared and step cubed")\n'
                "    return step_squared + step_cubed\n"
            ),
            "documentation": "",
            "version": None,
        },
        {
            "name": "steps",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": ["number_of_steps"],
            "optional_dependencies": [],
            "source": "def steps(number_of_steps: int) -> int:\n    return number_of_steps\n",
            "documentation": "",
            "version": None,
        },
        {
            "name": "sum_step_squared_plus_step_cubed",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": ["step_squared_plus_step_cubed"],
            "optional_dependencies": [],
            "source": (
                "def sum_step_squared_plus_step_cubed(step_squared_plus_step_cubed: int) -> int:\n"
                '    print("summing step squared")\n'
                "    return step_squared_plus_step_cubed\n"
            ),
            "documentation": "",
            "version": None,
        },
    ],
}


def test_export_execution():
    dr = driver.Builder().with_modules(no_parallel).build()
    graph = graph_types.HamiltonGraph.from_graph(dr.graph)

    # can't hard code HamiltonNode.version because it depends on Python version
    for n in graph.nodes:
        for json_n in EXPECTED_JSON["nodes"]:
            if json_n["name"] == n.name:
                json_n["version"] = n.version

    json_str = dr.export_execution(["final"])
    assert json_str == json.dumps(EXPECTED_JSON)



import os
import time
from datetime import datetime
from os import PathLike
from pathlib import Path
from typing import Any, Dict, Union
from urllib import parse

import pandas as pd

DATAFRAME_METADATA = "dataframe_metadata"
SQL_METADATA = "sql_metadata"
FILE_METADATA = "file_metadata"


def get_file_metadata(path: Union[str, Path, PathLike]) -> Dict[str, Any]:
    """Gives metadata from loading a file.

    Note: we reserve the right to change this schema. So if you're using this come
    chat so that we can make sure we don't break your code.

    This includes:
    - the file size
    - the file path
    - the last modified time
    - the current time
    """
    if isinstance(path, Path):
        path = str(path)
    parsed = parse.urlparse(path)
    size = None
    scheme = parsed.scheme
    last_modified = time.time()
    timestamp = datetime.now().utcnow().timestamp()
    notes = f"File metadata is unsupported for scheme: {scheme} or path: {path} does not exist."

    if parsed.scheme == "" and os.path.exists(path):
        size = os.path.getsize(path)
        last_modified = os.path.getmtime(path)
        notes = ""

    return {
        FILE_METADATA: {
            "size": size,
            "path": path,
            "last_modified": last_modified,
            "timestamp": timestamp,
            "scheme": scheme,
            "notes": notes,
        }
    }


def get_dataframe_metadata(df: pd.DataFrame) -> Dict[str, Any]:
    """Gives metadata from loading a dataframe.

    Note: we reserve the right to change this schema. So if you're using this come
    chat so that we can make sure we don't break your code.

    This includes:
    - the number of rows
    - the number of columns
    - the column names
    - the data types
    """
    metadata = {}
    try:
        metadata["rows"] = len(df)
    except TypeError:
        metadata["rows"] = None

    try:
        metadata["columns"] = len(df.columns)
    except (AttributeError, TypeError):
        metadata["columns"] = None

    try:
        metadata["column_names"] = list(df.columns)
    except (AttributeError, TypeError):
        metadata["column_names"] = None

    try:
        metadata["datatypes"] = [str(t) for t in list(df.dtypes)]
    except (AttributeError, TypeError):
        metadata["datatypes"] = None
    return {DATAFRAME_METADATA: metadata}


def get_file_and_dataframe_metadata(path: str, df: pd.DataFrame) -> Dict[str, Any]:
    """Gives metadata from loading a file and a dataframe.

    Note: we reserve the right to change this schema. So if you're using this come
    chat so that we can make sure we don't break your code.

    This includes:
        file_meta:
            - the file size
            - the file path
            - the last modified time
            - the current time
        dataframe_meta:
        - the number of rows
        - the number of columns
        - the column names
        - the data types
    """
    return {**get_file_metadata(path), **get_dataframe_metadata(df)}


def get_sql_metadata(query_or_table: str, results: Union[int, pd.DataFrame]) -> Dict[str, Any]:
    """Gives metadata from reading a SQL table or writing to SQL db.

    Note: we reserve the right to change this schema. So if you're using this come
    chat so that we can make sure we don't break your code.

    This includes:
    - the number of rows read, added, or to add.
    - the sql query (e.g., "SELECT foo FROM bar")
    - the table name (e.g., "bar")
    - the current time
    """
    query = query_or_table if "SELECT" in query_or_table else None
    table_name = query_or_table if "SELECT" not in query_or_table else None
    if isinstance(results, int):
        rows = results
    elif isinstance(results, pd.DataFrame):
        rows = len(results)
    else:
        rows = None
    return {
        SQL_METADATA: {
            "rows": rows,
            "query": query,
            "table_name": table_name,
            "timestamp": datetime.now().utcnow().timestamp(),
        }
    }



from hamilton.function_modifiers import parameterize_values


@parameterize_values(
    "param",
    {("parametrized_1", "doc"): 1, ("parametrized_2", "doc"): 2, ("parametrized_3", "doc"): 3},
)
def to_parametrize(param: int) -> int:
    """Function that should be parametrized to form multiple functions"""
    return param



"""
Module for dummy functions to test graph things with.
"""

# we import this to check we don't pull in this function when parsing this module.
from tests.resources import only_import_me


def A(b: int, c: int) -> int:
    """Function that should become part of the graph - A"""
    return b + c


def _do_not_import_me(some_input: int, some_input2: int) -> int:
    """Function that should not become part of the graph - _do_not_import_me."""
    only_import_me.this_is_not_something_we_should_import()
    return some_input + some_input2


def B(A: int) -> int:
    """Function that should become part of the graph - B"""
    return A * A


def C(A: int) -> int:  # empty string doc on purpose.
    return A * 2



"""This module exists so that people can download dataflows without the sf-hamilton-contrib package.

It will get clobbered when sf-hamilton-contrib is installed, which is good.
"""

import logging
from contextlib import contextmanager

__version__ = "__unknown__"  # this will be overwritten once sf-hamilton-contrib is installed.

from hamilton import telemetry


@contextmanager
def catch_import_errors(module_name: str, file_location: str, logger: logging.Logger):
    try:
        # Yield control to the inner block which will have the import statements.
        yield
        # After all imports succeed send telemetry
        if "." in module_name:
            telemetry.create_and_send_contrib_use(module_name, __version__)
        else:
            # we are importing it dynamically thus a "package" isn't present so file_location has the info.
            telemetry.create_and_send_contrib_use(file_location, __version__)
    except ImportError as e:
        location = file_location[: file_location.rfind("/")]
        logger.error("ImportError: %s", e)
        logger.error(
            "Please install the required packages. Options:\n"
            f"(1): with `pip install -r {location}/requirements.txt`\n"
        )
        raise e



from kedro.io import DataCatalog
from kedro.io.memory_dataset import MemoryDataset

from hamilton.plugins import kedro_extensions


def test_kedro_saver():
    dataset_name = "in_memory"
    data = 37
    catalog = DataCatalog({dataset_name: MemoryDataset()})

    saver = kedro_extensions.KedroSaver(dataset_name=dataset_name, catalog=catalog)
    saver.save_data(data)
    loaded_data = catalog.load(dataset_name)

    assert loaded_data == data


def test_kedro_loader():
    dataset_name = "in_memory"
    data = 37
    catalog = DataCatalog({dataset_name: MemoryDataset(data=data)})

    loader = kedro_extensions.KedroLoader(dataset_name=dataset_name, catalog=catalog)
    loaded_data, metadata = loader.load_data(int)

    assert loaded_data == data



import dataclasses
from typing import Any, Collection, Dict, Iterable, Literal, Optional, Sequence, Tuple, Type

import dlt
import pandas as pd
from dlt.common.destination.capabilities import TLoaderFileFormat
from dlt.common.schema import Schema, TColumnSchema

# importing TDestinationReferenceArg fails if Destination isn't imported
from dlt.extract.resource import DltResource

from hamilton import registry
from hamilton.io import utils
from hamilton.io.data_adapters import DataLoader, DataSaver

DATAFRAME_TYPES = [Iterable, pd.DataFrame]

# TODO add types for other Dataframe libraries
try:
    import pyarrow as pa

    DATAFRAME_TYPES.extend([pa.Table, pa.RecordBatch])
except ModuleNotFoundError:
    pass

# convert to tuple to dynamically define type `Union[DATAFRAME_TYPES]`
DATAFRAME_TYPES = tuple(DATAFRAME_TYPES)
COLUMN_FRIENDLY_DF_TYPE = False


@dataclasses.dataclass
class DltResourceLoader(DataLoader):
    resource: DltResource

    @classmethod
    def name(cls) -> str:
        return "dlt"

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [pd.DataFrame]

    def load_data(self, type_: Type) -> Tuple[pd.DataFrame, Dict[str, Any]]:
        """Creates a pipeline and conduct `extract` and `normalize` steps.
        Then, "load packages" are read with pandas
        """
        pipeline = dlt.pipeline(
            pipeline_name="Hamilton-DltResourceLoader", destination="filesystem"
        )
        pipeline.extract(self.resource)
        normalize_info = pipeline.normalize(loader_file_format="parquet")

        partition_file_paths = []
        package = normalize_info.load_packages[0]
        for job in package.jobs["new_jobs"]:
            if job.job_file_info.table_name == self.resource.name:
                partition_file_paths.append(job.file_path)

        # TODO use pyarrow directly to support different dataframe libraries
        # ref: https://github.com/dlt-hub/verified-sources/blob/master/sources/filesystem/readers.py
        # ref: https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetDataset.html#pyarrow.parquet.ParquetDataset
        df = pd.concat([pd.read_parquet(f) for f in partition_file_paths], ignore_index=True)

        # delete the pipeline
        pipeline.drop()

        metadata = utils.get_dataframe_metadata(df)
        return df, metadata


# TODO handle behavior with `combine=`, currently only supports materializing a single node
@dataclasses.dataclass
class DltDestinationSaver(DataSaver):
    """Materialize results using a dlt pipeline with the specified destination.

    In reference to an Extract, Transform, Load (ETL) pipeline, here, the Hamilton
    dataflow is responsible for Transform, and `DltDestination` for Load.
    """

    pipeline: dlt.Pipeline
    table_name: str
    primary_key: Optional[str] = None
    write_disposition: Optional[Literal["skip", "append", "replace", "merge"]] = None
    columns: Optional[Sequence[TColumnSchema]] = None
    schema: Optional[Schema] = None
    loader_file_format: Optional[TLoaderFileFormat] = None

    @classmethod
    def name(cls) -> str:
        return "dlt"

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return DATAFRAME_TYPES

    def _get_kwargs(self) -> dict:
        kwargs = {}
        fields_to_skip = ["pipeline"]
        for field in dataclasses.fields(self):
            field_value = getattr(self, field.name)
            if field.name in fields_to_skip:
                continue

            if field_value != field.default:
                kwargs[field.name] = field_value

        return kwargs

    # TODO get pyarrow table from polars, dask, etc.
    def save_data(self, data) -> Dict[str, Any]:
        """
        ref: https://dlthub.com/docs/dlt-ecosystem/verified-sources/arrow-pandas
        """
        if isinstance(data, dict):
            raise NotImplementedError(
                "DltDestinationSaver received data of type `dict`."
                "Currently, it doesn't support specifying `combine=base.DictResult()`"
            )

        load_info = self.pipeline.run(data, **self._get_kwargs())
        # follows the pattern of metadata output found in hamilton.io.utils
        return {"dlt_metadata": load_info.asdict()}


def register_data_loaders():
    """Function to register the data loaders for this extension."""
    for loader in [
        DltDestinationSaver,
        DltResourceLoader,
    ]:
        registry.register_adapter(loader)


register_data_loaders()



from typing import List

from hamilton.execution.grouping import NodeGroup, NodeGroupPurpose, TaskSpec

"""A set of utilities for debugging/printing out data"""
group_purpose_icons = {
    NodeGroupPurpose.EXPAND_UNORDERED: "⫳",
    NodeGroupPurpose.GATHER: "⋃",
    NodeGroupPurpose.EXECUTE_BLOCK: "᠅",
    NodeGroupPurpose.EXECUTE_SINGLE: "•",
}


def print_node_groups(node_groups: List[NodeGroup]):
    """Prints out the node groups in a clean, tree-like format.

    :param node_groups:
    :return:
    """
    for group in node_groups:
        node_icon = group_purpose_icons[group.purpose]
        print(f"{node_icon} {group.base_id}")
        for node_ in group.nodes:
            print(f"   • {node_.name} [ƒ({','.join(map(lambda n: n.name, node_.dependencies))})]")


def print_tasks(tasks: List[TaskSpec]):
    """Prints out the node groups in a clean, tree-like format.

    :param tasks:
    :return:
    """
    print()
    for task in tasks:
        node_icon = group_purpose_icons[task.purpose]
        print(f"{node_icon} {task.base_id} [ƒ({', '.join(task.base_dependencies)})]")
        for node_ in task.nodes:
            print(
                f"   • {node_.name}"
            )  # ƒ({', '.join(map(lambda n: n.name, node_.dependencies))})")



import pathlib
import sys

import polars as pl
import pytest
from polars.testing import assert_frame_equal

from hamilton.plugins.polars_extensions import (
    PolarsAvroReader,
    PolarsAvroWriter,
    PolarsCSVWriter,
    PolarsDatabaseReader,
    PolarsDatabaseWriter,
    PolarsFeatherWriter,
    PolarsJSONReader,
    PolarsJSONWriter,
    PolarsParquetWriter,
    PolarsSpreadsheetReader,
    PolarsSpreadsheetWriter,
)
from hamilton.plugins.polars_lazyframe_extensions import (
    PolarsScanCSVReader,
    PolarsScanFeatherReader,
    PolarsScanParquetReader,
)


@pytest.fixture
def df():
    yield pl.LazyFrame({"a": [1, 2], "b": [3, 4]})


def test_lazy_polars_lazyframe_csv(df: pl.LazyFrame, tmp_path: pathlib.Path) -> None:
    file = tmp_path / "test.csv"

    writer = PolarsCSVWriter(file=file)
    kwargs1 = writer._get_saving_kwargs()
    writer.save_data(df)

    reader = PolarsScanCSVReader(file=file)
    kwargs2 = reader._get_loading_kwargs()
    df2, metadata = reader.load_data(pl.LazyFrame)

    assert PolarsCSVWriter.applicable_types() == [pl.DataFrame, pl.LazyFrame]
    assert PolarsScanCSVReader.applicable_types() == [pl.LazyFrame]
    assert kwargs1["separator"] == ","
    assert kwargs2["has_header"] is True
    assert_frame_equal(df.collect(), df2.collect())


def test_lazy_polars_parquet(df: pl.LazyFrame, tmp_path: pathlib.Path) -> None:
    file = tmp_path / "test.parquet"

    writer = PolarsParquetWriter(file=file)
    kwargs1 = writer._get_saving_kwargs()
    writer.save_data(df)

    reader = PolarsScanParquetReader(file=file, n_rows=2)
    kwargs2 = reader._get_loading_kwargs()
    df2, metadata = reader.load_data(pl.LazyFrame)

    assert PolarsParquetWriter.applicable_types() == [pl.DataFrame, pl.LazyFrame]
    assert PolarsScanParquetReader.applicable_types() == [pl.LazyFrame]
    assert kwargs1["compression"] == "zstd"
    assert kwargs2["n_rows"] == 2
    assert_frame_equal(df.collect(), df2.collect())


def test_lazy_polars_feather(tmp_path: pathlib.Path) -> None:
    test_data_file_path = "tests/resources/data/test_load_from_data.feather"
    reader = PolarsScanFeatherReader(source=test_data_file_path)
    read_kwargs = reader._get_loading_kwargs()
    df, _ = reader.load_data(pl.LazyFrame)

    file_path = tmp_path / "test.dta"
    writer = PolarsFeatherWriter(file=file_path)
    write_kwargs = writer._get_saving_kwargs()
    metadata = writer.save_data(df.collect())

    assert PolarsScanFeatherReader.applicable_types() == [pl.LazyFrame]
    assert "n_rows" not in read_kwargs
    assert df.collect().shape == (4, 3)

    assert PolarsFeatherWriter.applicable_types() == [pl.DataFrame, pl.LazyFrame]
    assert "compression" in write_kwargs
    assert file_path.exists()
    assert metadata["file_metadata"]["path"] == str(file_path)
    assert metadata["dataframe_metadata"]["column_names"] == [
        "animal",
        "points",
        "environment",
    ]
    assert metadata["dataframe_metadata"]["datatypes"] == ["String", "Int64", "String"]


def test_lazy_polars_avro(df: pl.LazyFrame, tmp_path: pathlib.Path) -> None:
    file = tmp_path / "test.avro"

    writer = PolarsAvroWriter(file=file)
    kwargs1 = writer._get_saving_kwargs()
    writer.save_data(df)

    reader = PolarsAvroReader(file=file, n_rows=2)
    kwargs2 = reader._get_loading_kwargs()
    df2, metadata = reader.load_data(pl.DataFrame)

    assert PolarsAvroWriter.applicable_types() == [pl.DataFrame, pl.LazyFrame]
    assert PolarsAvroReader.applicable_types() == [pl.DataFrame]
    assert kwargs1["compression"] == "uncompressed"
    assert kwargs2["n_rows"] == 2
    assert_frame_equal(df.collect(), df2)


def test_polars_json(df: pl.LazyFrame, tmp_path: pathlib.Path) -> None:
    file = tmp_path / "test.json"
    writer = PolarsJSONWriter(file=file, pretty=True)
    kwargs1 = writer._get_saving_kwargs()
    writer.save_data(df)

    reader = PolarsJSONReader(source=file)
    kwargs2 = reader._get_loading_kwargs()
    df2, metadata = reader.load_data(pl.DataFrame)

    assert PolarsJSONWriter.applicable_types() == [pl.DataFrame, pl.LazyFrame]
    assert PolarsJSONReader.applicable_types() == [pl.DataFrame]
    assert kwargs1["pretty"]
    assert df2.shape == (2, 2)
    assert "schema" not in kwargs2
    assert_frame_equal(df.collect(), df2)


@pytest.mark.skipif(
    sys.version_info.major == 3 and sys.version_info.minor == 12,
    reason="weird connectorx error on 3.12",
)
def test_polars_database(df: pl.LazyFrame, tmp_path: pathlib.Path) -> None:
    conn = f"sqlite:///{tmp_path}/test.db"
    table_name = "test_table"

    writer = PolarsDatabaseWriter(table_name=table_name, connection=conn, if_table_exists="replace")
    kwargs1 = writer._get_saving_kwargs()
    writer.save_data(df)

    reader = PolarsDatabaseReader(query=f"SELECT * FROM {table_name}", connection=conn)
    kwargs2 = reader._get_loading_kwargs()
    df2, metadata = reader.load_data(pl.DataFrame)

    assert PolarsDatabaseWriter.applicable_types() == [pl.DataFrame, pl.LazyFrame]
    assert PolarsDatabaseReader.applicable_types() == [pl.DataFrame]
    assert kwargs1["if_table_exists"] == "replace"
    assert "batch_size" not in kwargs2
    assert df2.shape == (2, 2)
    assert_frame_equal(df.collect(), df2)


def test_polars_spreadsheet(df: pl.LazyFrame, tmp_path: pathlib.Path) -> None:
    file_path = tmp_path / "test.xlsx"
    writer = PolarsSpreadsheetWriter(workbook=file_path, worksheet="test_load_from_data_sheet")
    write_kwargs = writer._get_saving_kwargs()
    metadata = writer.save_data(df)

    reader = PolarsSpreadsheetReader(source=file_path, sheet_name="test_load_from_data_sheet")
    read_kwargs = reader._get_loading_kwargs()
    df2, _ = reader.load_data(pl.DataFrame)

    assert PolarsSpreadsheetWriter.applicable_types() == [pl.DataFrame, pl.LazyFrame]
    assert PolarsSpreadsheetReader.applicable_types() == [pl.DataFrame]
    assert file_path.exists()
    assert metadata["file_metadata"]["path"] == str(file_path)
    assert df.collect().shape == (2, 2)
    assert metadata["dataframe_metadata"]["column_names"] == ["a", "b"]
    assert metadata["dataframe_metadata"]["datatypes"] == ["Int64", "Int64"]
    assert_frame_equal(df.collect(), df2)
    assert "include_header" in write_kwargs
    assert write_kwargs["include_header"] is True
    assert "raise_if_empty" in read_kwargs
    assert read_kwargs["raise_if_empty"] is True
